'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

from haigha.writer import Writer
from haigha.frames.method_frame import MethodFrame
from haigha.classes.protocol_class import ProtocolClass


class ExchangeClass(ProtocolClass):

    '''
    Implements the AMQP Exchange class
    '''

    def __init__(self, *args, **kwargs):
        super(ExchangeClass, self).__init__(*args, **kwargs)
        self.dispatch_map = {
            11: self._recv_declare_ok,
            21: self._recv_delete_ok,
        }

        self._declare_cb = deque()
        self._delete_cb = deque()

    @property
    def name(self):
        return 'exchange'

    def _cleanup(self):
        '''
        Cleanup local data.
        '''
        self._declare_cb = None
        self._delete_cb = None
        super(ExchangeClass, self)._cleanup()

    def declare(self, exchange, type, passive=False, durable=False,
                nowait=True, arguments=None, ticket=None, cb=None):
        """
        Declare the exchange.

        exchange - The name of the exchange to declare
        type - One of
        """
        nowait = nowait and self.allow_nowait() and not cb

        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(exchange).\
            write_shortstr(type).\
            write_bits(passive, durable, False, False, nowait).\
            write_table(arguments or {})
        self.send_frame(MethodFrame(self.channel_id, 40, 10, args))

        if not nowait:
            self._declare_cb.append(cb)
            self.channel.add_synchronous_cb(self._recv_declare_ok)

    def _recv_declare_ok(self, _method_frame):
        '''
        Confirmation that exchange was declared.
        '''
        cb = self._declare_cb.popleft()
        if cb:
            cb()

    def delete(self, exchange, if_unused=False, nowait=True, ticket=None,
               cb=None):
        '''
        Delete an exchange.
        '''
        nowait = nowait and self.allow_nowait() and not cb

        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(exchange).\
            write_bits(if_unused, nowait)
        self.send_frame(MethodFrame(self.channel_id, 40, 20, args))

        if not nowait:
            self._delete_cb.append(cb)
            self.channel.add_synchronous_cb(self._recv_delete_ok)

    def _recv_delete_ok(self, _method_frame):
        '''
        Confirmation that exchange was deleted.
        '''
        cb = self._delete_cb.popleft()
        if cb:
            cb()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque


class ChannelPool(object):

    '''
    Manages a pool of channels for transaction-based publishing.  This allows a
    client to use as many channels as are necessary to publish while not
    creating a backlog of transactions that slows throughput and consumes
    memory.

    The pool can accept an optional `size` argument in the ctor, which caps
    the number of channels which the pool will allocate. If no channels are
    available on `publish()`, the message will be locally queued and sent as
    soon as a channel is available. It is recommended that you use the pool
    with a max size, as each channel consumes memory on the broker and it is
    possible to exercise memory limit protection seems on the broker due to
    number of channels.
    '''

    def __init__(self, connection, size=None):
        '''Initialize the channel on a connection.'''
        self._connection = connection
        self._free_channels = set()
        self._size = size
        self._queue = deque()
        self._channels = 0

    def publish(self, *args, **kwargs):
        '''
        Publish a message. Caller can supply an optional callback which will
        be fired when the transaction is committed. Tries very hard to avoid
        closed and inactive channels, but a ChannelError or ConnectionError
        may still be raised.
        '''
        user_cb = kwargs.pop('cb', None)

        # If the first channel we grab is inactive, continue fetching until
        # we get an active channel, then put the inactive channels back in
        # the pool. Try to keep the overhead to a minimum.
        channel = self._get_channel()

        if channel and not channel.active:
            inactive_channels = set()
            while channel and not channel.active:
                inactive_channels.add(channel)
                channel = self._get_channel()
            self._free_channels.update(inactive_channels)

        # When the transaction is committed, add the channel back to the pool
        # and call any user-defined callbacks. If there is anything in queue,
        # pop it and call back to publish(). Only do so if the channel is
        # still active though, because otherwise the message will end up at
        # the back of the queue, breaking the original order.
        def committed():
            self._free_channels.add(channel)
            if channel.active and not channel.closed:
                self._process_queue()
            if user_cb is not None:
                user_cb()

        if channel:
            channel.publish_synchronous(*args, cb=committed, **kwargs)
        else:
            kwargs['cb'] = user_cb
            self._queue.append((args, kwargs))

    def _process_queue(self):
        '''
        If there are any message in the queue, process one of them.
        '''
        if len(self._queue):
            args, kwargs = self._queue.popleft()
            self.publish(*args, **kwargs)

    def _get_channel(self):
        '''
        Fetch a channel from the pool. Will return a new one if necessary. If
        a channel in the free pool is closed, will remove it. Will return None
        if we hit the cap. Will clean up any channels that were published to
        but closed due to error.
        '''
        while len(self._free_channels):
            rval = self._free_channels.pop()
            if not rval.closed:
                return rval
            # don't adjust _channels value because the callback will do that
            # and we don't want to double count it.

        if not self._size or self._channels < self._size:
            rval = self._connection.channel()
            self._channels += 1
            rval.add_close_listener(self._channel_closed_cb)
            return rval

    def _channel_closed_cb(self, channel):
        '''
        Callback when channel closes.
        '''
        self._channels -= 1
        self._process_queue()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

from haigha.message import Message
from haigha.writer import Writer
from haigha.frames.method_frame import MethodFrame
from haigha.frames.header_frame import HeaderFrame
from haigha.frames.content_frame import ContentFrame
from haigha.classes.protocol_class import ProtocolClass


class BasicClass(ProtocolClass):

    '''
    Implements the AMQP Basic class
    '''

    def __init__(self, *args, **kwargs):
        super(BasicClass, self).__init__(*args, **kwargs)
        self.dispatch_map = {
            11: self._recv_qos_ok,
            21: self._recv_consume_ok,
            31: self._recv_cancel_ok,
            50: self._recv_return,
            60: self._recv_deliver,
            71: self._recv_get_response,   # see impl
            72: self._recv_get_response,   # see impl
            111: self._recv_recover_ok,
        }

        self._consumer_tag_id = 0
        self._pending_consumers = deque()
        self._consumer_cb = {}
        self._get_cb = deque()
        self._recover_cb = deque()
        self._cancel_cb = deque()

    @property
    def name(self):
        return 'basic'

    def _cleanup(self):
        '''
        Cleanup all the local data.
        '''
        self._pending_consumers = None
        self._consumer_cb = None
        self._get_cb = None
        self._recover_cb = None
        self._cancel_cb = None
        super(BasicClass, self)._cleanup()

    def _generate_consumer_tag(self):
        '''
        Generate the next consumer tag.

        The consumer tag is local to a channel, so two clients can use the
        same consumer tags.
        '''
        self._consumer_tag_id += 1
        return "channel-%d-%d" % (self.channel_id, self._consumer_tag_id)

    def qos(self, prefetch_size=0, prefetch_count=0, is_global=False):
        '''
        Set QoS on this channel.
        '''
        args = Writer()
        args.write_long(prefetch_size).\
            write_short(prefetch_count).\
            write_bit(is_global)
        self.send_frame(MethodFrame(self.channel_id, 60, 10, args))

        self.channel.add_synchronous_cb(self._recv_qos_ok)

    def _recv_qos_ok(self, _method_frame):
        # No arguments, nothing to do
        pass

    def consume(self, queue, consumer, consumer_tag='', no_local=False,
                no_ack=True, exclusive=False, nowait=True, ticket=None,
                cb=None):
        '''
        Start a queue consumer. If `cb` is supplied, will be called when
        broker confirms that consumer is registered.
        '''
        nowait = nowait and self.allow_nowait() and not cb

        if nowait and consumer_tag == '':
            consumer_tag = self._generate_consumer_tag()

        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(queue).\
            write_shortstr(consumer_tag).\
            write_bits(no_local, no_ack, exclusive, nowait).\
            write_table({})  # unused according to spec
        self.send_frame(MethodFrame(self.channel_id, 60, 20, args))

        if not nowait:
            self._pending_consumers.append((consumer, cb))
            self.channel.add_synchronous_cb(self._recv_consume_ok)
        else:
            self._consumer_cb[consumer_tag] = consumer

    def _recv_consume_ok(self, method_frame):
        consumer_tag = method_frame.args.read_shortstr()
        consumer, cb = self._pending_consumers.popleft()

        self._consumer_cb[consumer_tag] = consumer
        if cb:
            cb()

    def cancel(self, consumer_tag='', nowait=True, consumer=None, cb=None):
        '''
        Cancel a consumer. Can choose to delete based on a consumer tag or
        the function which is consuming.  If deleting by function, take care
        to only use a consumer once per channel.
        '''
        if consumer:
            for (tag, func) in self._consumer_cb.iteritems():
                if func == consumer:
                    consumer_tag = tag
                    break

        nowait = nowait and self.allow_nowait() and not cb

        args = Writer()
        args.write_shortstr(consumer_tag).\
            write_bit(nowait)
        self.send_frame(MethodFrame(self.channel_id, 60, 30, args))

        if not nowait:
            self._cancel_cb.append(cb)
            self.channel.add_synchronous_cb(self._recv_cancel_ok)
        else:
            try:
                del self._consumer_cb[consumer_tag]
            except KeyError:
                self.logger.warning(
                    'no callback registered for consumer tag " %s "',
                    consumer_tag)

    def _recv_cancel_ok(self, method_frame):
        consumer_tag = method_frame.args.read_shortstr()
        try:
            del self._consumer_cb[consumer_tag]
        except KeyError:
            self.logger.warning(
                'no callback registered for consumer tag " %s "', consumer_tag)

        cb = self._cancel_cb.popleft()
        if cb:
            cb()

    def publish(self, msg, exchange, routing_key, mandatory=False,
                immediate=False, ticket=None):
        '''
        publish a message.
        '''
        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(exchange).\
            write_shortstr(routing_key).\
            write_bits(mandatory, immediate)

        self.send_frame(MethodFrame(self.channel_id, 60, 40, args))
        self.send_frame(
            HeaderFrame(self.channel_id, 60, 0, len(msg), msg.properties))

        f_max = self.channel.connection.frame_max
        for f in ContentFrame.create_frames(self.channel_id, msg.body, f_max):
            self.send_frame(f)

    def return_msg(self, reply_code, reply_text, exchange, routing_key):
        '''
        Return a failed message.  Not named "return" because python interpreter
        can't deal with that.
        '''
        args = Writer()
        args.write_short(reply_code).\
            write_shortstr(reply_text).\
            write_shortstr(exchange).\
            write_shortstr(routing_key)

        self.send_frame(MethodFrame(self.channel_id, 60, 50, args))

    def _recv_return(self, _method_frame):
        # This seems like the right place to callback that the operation has
        # completed.
        pass

    def _recv_deliver(self, method_frame):
        msg = self._read_msg(method_frame,
                             with_consumer_tag=True, with_message_count=False)

        func = self._consumer_cb.get(msg.delivery_info['consumer_tag'], None)
        if func:
            func(msg)

    def get(self, queue, consumer=None, no_ack=True, ticket=None):
        '''
        Ask to fetch a single message from a queue.  If a consumer is supplied,
        the consumer will be called with either a Message argument, or None if
        there is no message in queue. If a synchronous transport, Message or
        None is returned.
        '''
        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(queue).\
            write_bit(no_ack)

        self._get_cb.append(consumer)
        self.send_frame(MethodFrame(self.channel_id, 60, 70, args))
        return self.channel.add_synchronous_cb(self._recv_get_response)

    def _recv_get_response(self, method_frame):
        '''
        Handle either get_ok or get_empty. This is a hack because the
        synchronous callback stack is expecting one method to satisfy the
        expectation. To keep that loop as tight as possible, work within
        those constraints. Use of get is not recommended anyway.
        '''
        if method_frame.method_id == 71:
            return self._recv_get_ok(method_frame)
        elif method_frame.method_id == 72:
            return self._recv_get_empty(method_frame)

    def _recv_get_ok(self, method_frame):
        msg = self._read_msg(method_frame,
                             with_consumer_tag=False, with_message_count=True)
        cb = self._get_cb.popleft()
        if cb:
            cb(msg)
        return msg

    def _recv_get_empty(self, _method_frame):
        # On empty, call back with None as the argument so that user code knows
        # it's empty and can take next action
        cb = self._get_cb.popleft()
        if cb:
            cb(None)

    def ack(self, delivery_tag, multiple=False):
        '''
        Acknowledge delivery of a message.  If multiple=True, acknowledge up-to
        and including delivery_tag.
        '''
        args = Writer()
        args.write_longlong(delivery_tag).\
            write_bit(multiple)

        self.send_frame(MethodFrame(self.channel_id, 60, 80, args))

    def reject(self, delivery_tag, requeue=False):
        '''
        Reject a message.
        '''
        args = Writer()
        args.write_longlong(delivery_tag).\
            write_bit(requeue)

        self.send_frame(MethodFrame(self.channel_id, 60, 90, args))

    def recover_async(self, requeue=False):
        '''
        Redeliver all unacknowledged messages on this channel.

        This method is deprecated in favour of the synchronous
        recover/recover-ok
        '''
        args = Writer()
        args.write_bit(requeue)

        self.send_frame(MethodFrame(self.channel_id, 60, 100, args))

    def recover(self, requeue=False, cb=None):
        '''
        Ask server to redeliver all unacknowledged messages.
        '''
        args = Writer()
        args.write_bit(requeue)

        # The XML spec is incorrect; this method is always synchronous
        #  http://lists.rabbitmq.com/pipermail/rabbitmq-discuss/2011-January/010738.html
        self._recover_cb.append(cb)
        self.send_frame(MethodFrame(self.channel_id, 60, 110, args))
        self.channel.add_synchronous_cb(self._recv_recover_ok)

    def _recv_recover_ok(self, _method_frame):
        cb = self._recover_cb.popleft()
        if cb:
            cb()

    def _read_msg(self, method_frame, with_consumer_tag=False,
                  with_message_count=False):
        '''
        Support method to read a Message from the current frame buffer.
        Will return a Message, or re-queue current frames and raise a
        FrameUnderflow. Takes an optional argument on whether to read the
        consumer tag so it can be used for both deliver and get-ok.
        '''
        # No need to assert that is instance of Header or Content frames
        # because failure to access as such will result in exception that
        # channel will pick up and handle accordingly.
        header_frame = self.channel.next_frame()
        if header_frame:
            size = header_frame.size
            body = bytearray()
            rbuf_frames = deque([header_frame, method_frame])

            while len(body) < size:
                content_frame = self.channel.next_frame()
                if content_frame:
                    rbuf_frames.appendleft(content_frame)
                    body.extend(content_frame.payload.buffer())
                else:
                    self.channel.requeue_frames(rbuf_frames)
                    raise self.FrameUnderflow()
        else:
            self.channel.requeue_frames([method_frame])
            raise self.FrameUnderflow()

        if with_consumer_tag:
            consumer_tag = method_frame.args.read_shortstr()
        delivery_tag = method_frame.args.read_longlong()
        redelivered = method_frame.args.read_bit()
        exchange = method_frame.args.read_shortstr()
        routing_key = method_frame.args.read_shortstr()
        if with_message_count:
            message_count = method_frame.args.read_long()

        delivery_info = {
            'channel': self.channel,
            'delivery_tag': delivery_tag,
            'redelivered': redelivered,
            'exchange': exchange,
            'routing_key': routing_key,
        }
        if with_consumer_tag:
            delivery_info['consumer_tag'] = consumer_tag
        if with_message_count:
            delivery_info['message_count'] = message_count

        return Message(body=body, delivery_info=delivery_info,
                       **header_frame.properties)



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.frames.frame import Frame
from haigha.writer import Writer


class HeartbeatFrame(Frame):

    '''
    Frame for heartbeats.
    '''

    @classmethod
    def type(cls):
        # NOTE: The PDF spec say this should be 4 but the xml spec say it
        # should be 8 RabbitMQ seems to implement this as 8, but maybe that's
        # a difference between 0.8 and 0.9 protocols
        # Using Rabbit 2.1.1 and protocol 0.9.1 it seems that 8 is indeed
        # the correct type @AW
        # PDF spec: http://www.amqp.org/confluence/download/attachments/720900/amqp0-9-1.pdf?version=1&modificationDate=1227526523000
        # XML spec: http://www.amqp.org/confluence/download/attachments/720900/amqp0-9-1.xml?version=1&modificationDate=1227526672000
        # This is addressed in
        # http://dev.rabbitmq.com/wiki/Amqp091Errata#section_29
        return 8

    @classmethod
    def parse(self, channel_id, payload):
        return HeartbeatFrame(channel_id)

    def write_frame(self, buf):
        writer = Writer(buf)
        writer.write_octet(self.type())
        writer.write_short(self.channel_id)
        writer.write_long(0)
        writer.write_octet(0xce)

HeartbeatFrame.register()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''


class ProtocolClass(object):

    '''
    The base class of all protocol classes.
    '''

    class ProtocolError(Exception):
        pass

    class InvalidMethod(ProtocolError):
        pass

    class FrameUnderflow(ProtocolError):
        pass

    dispatch_map = {}

    def __init__(self, channel):
        '''
        Construct this protocol class on a channel.
        '''
        # Cache the channel id so that cleanup can remove the circular channel
        # reference but id is still accessible (it's useful!)
        self._channel = channel
        self._channel_id = channel.channel_id

    @property
    def channel(self):
        return self._channel

    @property
    def channel_id(self):
        return self._channel_id

    @property
    def logger(self):
        return self._channel.logger

    @property
    def default_ticket(self):
        return 0

    @property
    def name(self):
        '''The name given this in the protocol, i.e. 'basic', 'tx', etc'''
        raise NotImplementedError('must provide a name for %s' % (self))

    def allow_nowait(self):
        '''
        Return True if the transport or  channel allows nowait,
        False otherwise.
        '''
        return not self._channel.synchronous

    def _cleanup(self):
        '''
        "Private" call from Channel when it's shutting down so that local
        data can be cleaned up and references closed out. It's strongly
        recommended that subclasses call this /after/ doing their own cleanup .
        Note that this removes reference to both the channel and the dispatch
        map.
        '''
        self._channel = None
        self.dispatch_map = None

    def dispatch(self, method_frame):
        '''
        Dispatch a method for this protocol.
        '''
        method = self.dispatch_map.get(method_frame.method_id)
        if method:
            callback = self.channel.clear_synchronous_cb(method)
            callback(method_frame)
        else:
            raise self.InvalidMethod(
                "no method is registered with id: %d" % method_frame.method_id)

    def send_frame(self, frame):
        '''
        Send a frame
        '''
        self.channel.send_frame(frame)



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.frames.method_frame import MethodFrame
from haigha.classes.protocol_class import ProtocolClass

from collections import deque


class TransactionClass(ProtocolClass):

    '''
    Implements the AMQP Transaction class
    '''

    class TransactionsNotEnabled(ProtocolClass.ProtocolError):

        '''Tried to use transactions without enabling them.'''

    def __init__(self, *args, **kwargs):
        super(TransactionClass, self).__init__(*args, **kwargs)
        self.dispatch_map = {
            11: self._recv_select_ok,
            21: self._recv_commit_ok,
            31: self._recv_rollback_ok,
        }

        self._enabled = False
        self._select_cb = deque()
        self._commit_cb = deque()
        self._rollback_cb = deque()

    @property
    def name(self):
        return 'tx'

    @property
    def enabled(self):
        '''Get whether transactions have been enabled.'''
        return self._enabled

    def _cleanup(self):
        '''
        Cleanup all the local data.
        '''
        self._select_cb = None
        self._commit_cb = None
        self._rollback_cb = None
        super(TransactionClass, self)._cleanup()

    def select(self, cb=None):
        '''
        Set this channel to use transactions.
        '''
        if not self._enabled:
            self._enabled = True
            self.send_frame(MethodFrame(self.channel_id, 90, 10))
            self._select_cb.append(cb)
            self.channel.add_synchronous_cb(self._recv_select_ok)

    def _recv_select_ok(self, _method_frame):
        cb = self._select_cb.popleft()
        if cb:
            cb()

    def commit(self, cb=None):
        '''
        Commit the current transaction.  Caller can specify a callback to use
        when the transaction is committed.
        '''
        # Could call select() but spec 1.9.2.3 says to raise an exception
        if not self.enabled:
            raise self.TransactionsNotEnabled()

        self.send_frame(MethodFrame(self.channel_id, 90, 20))
        self._commit_cb.append(cb)
        self.channel.add_synchronous_cb(self._recv_commit_ok)

    def _recv_commit_ok(self, _method_frame):
        cb = self._commit_cb.popleft()
        if cb:
            cb()

    def rollback(self, cb=None):
        '''
        Abandon all message publications and acks in the current transaction.
        Caller can specify a callback to use when the transaction has been
        aborted.
        '''
        # Could call select() but spec 1.9.2.5 says to raise an exception
        if not self.enabled:
            raise self.TransactionsNotEnabled()

        self.send_frame(MethodFrame(self.channel_id, 90, 30))
        self._rollback_cb.append(cb)
        self.channel.add_synchronous_cb(self._recv_rollback_ok)

    def _recv_rollback_ok(self, _method_frame):
        cb = self._rollback_cb.popleft()
        if cb:
            cb()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from struct import Struct
from datetime import datetime
from decimal import Decimal


class Reader(object):

    """
    A stream-like reader object that supports all the basic data types of AMQP.
    """

    class ReaderError(Exception):

        '''Base class for all reader errors.'''
    class BufferUnderflow(ReaderError):

        '''Not enough bytes to satisfy the request.'''
    class FieldError(ReaderError):

        '''Unsupported field type was read.'''

    def __init__(self, source, start_pos=0, size=None):
        """
        source should be a bytearray, io object with a read() method, another
        Reader, a plain or unicode string. Can be allocated over a slice
        of source.
        """
        # Note: buffer used here because unpack_from can't accept an array,
        # which I think is related to http://bugs.python.org/issue7827
        if isinstance(source, bytearray):
            self._input = buffer(source)
        elif isinstance(source, Reader):
            self._input = source._input
        elif hasattr(source, 'read'):
            self._input = buffer(source.read())
        elif isinstance(source, str):
            self._input = buffer(source)
        elif isinstance(source, unicode):
            self._input = buffer(source.encode('utf8'))
        else:
            raise ValueError(
                'Reader needs a bytearray, io object or plain string')

        self._start_pos = self._pos = start_pos
        self._end_pos = len(self._input)
        if size:
            self._end_pos = self._start_pos + size

    def __str__(self):
        return ''.join(['\\x%s' % (c.encode('hex')) for c in
                       self._input[self._start_pos:self._end_pos]])

    def tell(self):
        '''
        Current position
        '''
        return self._pos

    def seek(self, offset, whence=0):
        '''
        Simple seek. Follows standard interface.
        '''
        if whence == 0:
            self._pos = self._start_pos + offset
        elif whence == 1:
            self._pos += offset
        else:
            self._pos = (self._end_pos - 1) + offset

    def _check_underflow(self, n):
        '''
        Raise BufferUnderflow if there's not enough bytes to satisfy
        the request.
        '''
        if self._pos + n > self._end_pos:
            raise self.BufferUnderflow()

    def __len__(self):
        '''
        Supports content framing in Channel
        '''
        return self._end_pos - self._start_pos

    def buffer(self):
        '''
        Get a copy of the buffer that this is reading from. Returns a
        buffer object
        '''
        return buffer(self._input, self._start_pos,
                      (self._end_pos - self._start_pos))

    def read(self, n):
        """
        Read n bytes.

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        """
        self._check_underflow(n)
        rval = self._input[self._pos:self._pos + n]
        self._pos += n
        return rval

    def read_bit(self):
        """
        Read a single boolean value, returns 0 or 1. Convience for single
        bit fields.

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        """
        # Perform a faster check on underflow
        if self._pos >= self._end_pos:
            raise self.BufferUnderflow()
        result = ord(self._input[self._pos]) & 1
        self._pos += 1
        return result

    def read_bits(self, num):
        '''
        Read several bits packed into the same field. Will return as a list.
        The bit field itself is little-endian, though the order of the
        returned array looks big-endian for ease of decomposition.

        Reader('\x02').read_bits(2) -> [False,True]
        Reader('\x08').read_bits(2) ->
            [False,True,False,False,False,False,False,False]
        first_field, second_field = Reader('\x02').read_bits(2)

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        Will raise ValueError if num < 0 or num > 9
        '''
        # Perform a faster check on underflow
        if self._pos >= self._end_pos:
            raise self.BufferUnderflow()
        if num < 0 or num >= 9:
            raise ValueError("8 bits per field")
        field = ord(self._input[self._pos])
        result = map(lambda x: field >> x & 1, xrange(num))
        self._pos += 1
        return result

    def read_octet(self, unpacker=Struct('B').unpack_from,
                   size=Struct('B').size):
        """
        Read one byte, return as an integer

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        Will raise struct.error if the data is malformed
        """
        # Technically should look at unpacker.size, but skipping that is way
        # faster and this method is the most-called of the readers
        if self._pos >= self._end_pos:
            raise self.BufferUnderflow()
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def read_short(self, unpacker=Struct('>H').unpack_from,
                   size=Struct('>H').size):
        """
        Read an unsigned 16-bit integer

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        Will raise struct.error if the data is malformed
        """
        self._check_underflow(size)
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def read_long(self, unpacker=Struct('>I').unpack_from,
                  size=Struct('>I').size):
        """
        Read an unsigned 32-bit integer

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        Will raise struct.error if the data is malformed
        """
        self._check_underflow(size)
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def read_longlong(self, unpacker=Struct('>Q').unpack_from,
                      size=Struct('>Q').size):
        """
        Read an unsigned 64-bit integer

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        Will raise struct.error if the data is malformed
        """
        self._check_underflow(size)
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def read_shortstr(self):
        """
        Read a utf-8 encoded string that's stored in up to
        255 bytes.

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        Will raise UnicodeDecodeError if the text is mal-formed.
        Will raise struct.error if the data is malformed
        """
        slen = self.read_octet()
        return self.read(slen)

    def read_longstr(self):
        """
        Read a string that's up to 2**32 bytes, the encoding
        isn't specified in the AMQP spec, so just return it as
        a plain Python string.

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        Will raise struct.error if the data is malformed
        """
        slen = self.read_long()
        return self.read(slen)

    def read_timestamp(self):
        """
        Read and AMQP timestamp, which is a 64-bit integer representing
        seconds since the Unix epoch in 1-second resolution.  Return as
        a Python datetime.datetime object, expressed as UTC time.

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        Will raise struct.error if the data is malformed
        """
        return datetime.utcfromtimestamp(self.read_longlong())

    def read_table(self):
        """
        Read an AMQP table, and return as a Python dictionary.

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        Will raise UnicodeDecodeError if the text is mal-formed.
        Will raise struct.error if the data is malformed
        """
        # Only need to check underflow on the table once
        tlen = self.read_long()
        self._check_underflow(tlen)
        end_pos = self._pos + tlen
        result = {}
        while self._pos < end_pos:
            name = self._field_shortstr()
            result[name] = self._read_field()
        return result

    def _read_field(self):
        '''
        Read a single byte for field type, then read the value.
        '''
        ftype = self._input[self._pos]
        self._pos += 1

        reader = self.field_type_map.get(ftype)
        if reader:
            return reader(self)

        raise Reader.FieldError('Unknown field type %s', ftype)

    def _field_bool(self):
        result = ord(self._input[self._pos]) & 1
        self._pos += 1
        return result

    def _field_short_short_int(self, unpacker=Struct('b').unpack_from,
                               size=Struct('b').size):
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def _field_short_short_uint(self, unpacker=Struct('B').unpack_from,
                                size=Struct('B').size):
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def _field_short_int(self, unpacker=Struct('>h').unpack_from,
                         size=Struct('>h').size):
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def _field_short_uint(self, unpacker=Struct('>H').unpack_from,
                          size=Struct('>H').size):
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def _field_long_int(self, unpacker=Struct('>i').unpack_from,
                        size=Struct('>i').size):
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def _field_long_uint(self, unpacker=Struct('>I').unpack_from,
                         size=Struct('>I').size):
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def _field_long_long_int(self, unpacker=Struct('>q').unpack_from,
                             size=Struct('>q').size):
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def _field_long_long_uint(self, unpacker=Struct('>Q').unpack_from,
                              size=Struct('>Q').size):
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def _field_float(self, unpacker=Struct('>f').unpack_from,
                     size=Struct('>f').size):
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    def _field_double(self, unpacker=Struct('>d').unpack_from,
                      size=Struct('>d').size):
        rval = unpacker(self._input, self._pos)[0]
        self._pos += size
        return rval

    # Coding to http://dev.rabbitmq.com/wiki/Amqp091Errata#section_3 which
    # differs from spec in that the value is signed.
    def _field_decimal(self):
        d = self._field_short_short_uint()
        n = self._field_long_int()
        return Decimal(n) / Decimal(10 ** d)

    def _field_shortstr(self):
        slen = self._field_short_short_uint()
        rval = self._input[self._pos:self._pos + slen]
        self._pos += slen
        return rval

    def _field_longstr(self):
        slen = self._field_long_uint()
        rval = self._input[self._pos:self._pos + slen]
        self._pos += slen
        return rval

    def _field_array(self):
        alen = self.read_long()
        end_pos = self._pos + alen
        rval = []
        while self._pos < end_pos:
            rval.append(self._read_field())
        return rval

    def _field_timestamp(self):
        """
        Read and AMQP timestamp, which is a 64-bit integer representing
        seconds since the Unix epoch in 1-second resolution.  Return as
        a Python datetime.datetime object, expressed as UTC time.

        Will raise BufferUnderflow if there's not enough bytes in the buffer.
        Will raise struct.error if the data is malformed
        """
        return datetime.utcfromtimestamp(self._field_long_long_uint())

    def _field_bytearray(self):
        slen = self._field_long_uint()
        rval = bytearray(self._input[self._pos:self._pos + slen])
        self._pos += slen
        return rval

    def _field_none(self):
        return None

    # A mapping for quick lookups
    # Rabbit and Qpid 0.9.1 mapping
    # Based on: http://www.rabbitmq.com/amqp-0-9-1-errata.html (3. Field types)
    field_type_map = {
        't': _field_bool,
        'b': _field_short_short_int,
        's': _field_short_int,
        'I': _field_long_int,
        'l': _field_long_long_int,
        'f': _field_float,
        'd': _field_double,
        'D': _field_decimal,
        'S': _field_longstr,
        'A': _field_array,
        'T': _field_timestamp,
        'F': read_table,
        'V': _field_none,
        'x': _field_bytearray,
    }

    # 0.9.1 spec mapping
    #  field_type_map = {
    #   't' : _field_bool,
    #   'b' : _field_short_short_int,
    #   'B' : _field_short_short_uint,
    #   'U' : _field_short_int,
    #   'u' : _field_short_uint,
    #   'I' : _field_long_int,
    #   'i' : _field_long_uint,
    #   'L' : _field_long_long_int,
    #   'l' : _field_long_long_uint,
    #   'f' : _field_float,
    #   'd' : _field_double,
    #   'D' : _field_decimal,
    #   's' : _field_shortstr,
    #   'S' : _field_longstr,
    #   'A' : _field_array,
    #   'T' : _field_timestamp,
    #   'F' : read_table,
    #   'V' : _field_none,
    # }



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from struct import Struct
from calendar import timegm
from datetime import datetime
from decimal import Decimal
from operator import xor


class Writer(object):

    """
    Implements writing of structured AMQP data. Buffers data directly to a
    bytearray or a buffer supplied in the constructor. The buffer must
    supply append, extend and struct.pack_into semantics.
    """

    def __init__(self, buf=None):
        if buf is not None:
            self._output_buffer = buf
        else:
            self._output_buffer = bytearray()

    def __str__(self):
        return ''.join([
            '\\x%s' % (chr(c).encode('hex')) for c in self._output_buffer])

    __repr__ = __str__

    def __eq__(self, other):
        if isinstance(other, Writer):
            return self._output_buffer == other._output_buffer
        return False

    def buffer(self):
        '''
        Get the buffer that this has written to. Returns bytearray.
        '''
        return self._output_buffer

    def write(self, s):
        """
        Write a plain Python string, with no special encoding.
        """
        self._output_buffer.extend(s)
        return self

    def write_bits(self, *args):
        '''
        Write multiple bits in a single byte field. The bits will be written in
        little-endian order, but should be supplied in big endian order. Will
        raise ValueError when more than 8 arguments are supplied.

        write_bits(True, False) => 0x02
        '''
        # Would be nice to make this a bit smarter
        if len(args) > 8:
            raise ValueError("Can only write 8 bits at a time")

        self._output_buffer.append(chr(
            reduce(lambda x, y: xor(x, args[y] << y), xrange(len(args)), 0)))

        return self

    def write_bit(self, b, pack=Struct('B').pack):
        '''
        Write a single bit. Convenience method for single bit args.
        '''
        self._output_buffer.append(pack(True if b else False))
        return self

    def write_octet(self, n, pack=Struct('B').pack):
        """
        Write an integer as an unsigned 8-bit value.
        """
        if 0 <= n <= 255:
            self._output_buffer.append(pack(n))
        else:
            raise ValueError('Octet %d out of range 0..255', n)
        return self

    def write_short(self, n, pack=Struct('>H').pack):
        """
        Write an integer as an unsigned 16-bit value.
        """
        if 0 <= n <= 0xFFFF:
            self._output_buffer.extend(pack(n))
        else:
            raise ValueError('Short %d out of range 0..0xFFFF', n)
        return self

    def write_short_at(self, n, pos, pack_into=Struct('>H').pack_into):
        '''
        Write an unsigned 16bit value at a specific position in the buffer.
        Used for writing tables and frames.
        '''
        if 0 <= n <= 0xFFFF:
            pack_into(self._output_buffer, pos, n)
        else:
            raise ValueError('Short %d out of range 0..0xFFFF', n)
        return self

    def write_long(self, n, pack=Struct('>I').pack):
        """
        Write an integer as an unsigned 32-bit value.
        """
        if 0 <= n <= 0xFFFFFFFF:
            self._output_buffer.extend(pack(n))
        else:
            raise ValueError('Long %d out of range 0..0xFFFFFFFF', n)
        return self

    def write_long_at(self, n, pos, pack_into=Struct('>I').pack_into):
        '''
        Write an unsigned 32bit value at a specific position in the buffer.
        Used for writing tables and frames.
        '''
        if 0 <= n <= 0xFFFFFFFF:
            pack_into(self._output_buffer, pos, n)
        else:
            raise ValueError('Long %d out of range 0..0xFFFFFFFF', n)
        return self

    def write_longlong(self, n, pack=Struct('>Q').pack):
        """
        Write an integer as an unsigned 64-bit value.
        """
        if 0 <= n <= 0xFFFFFFFFFFFFFFFF:
            self._output_buffer.extend(pack(n))
        else:
            raise ValueError(
                'Longlong %d out of range 0..0xFFFFFFFFFFFFFFFF', n)
        return self

    def write_shortstr(self, s):
        """
        Write a string up to 255 bytes long after encoding.  If passed
        a unicode string, encode as UTF-8.
        """
        if isinstance(s, unicode):
            s = s.encode('utf-8')
        self.write_octet(len(s))
        self.write(s)
        return self

    def write_longstr(self, s):
        """
        Write a string up to 2**32 bytes long after encoding.  If passed
        a unicode string, encode as UTF-8.
        """
        if isinstance(s, unicode):
            s = s.encode('utf-8')
        self.write_long(len(s))
        self.write(s)
        return self

    def write_timestamp(self, t, pack=Struct('>Q').pack):
        """
        Write out a Python datetime.datetime object as a 64-bit integer
        representing seconds since the Unix UTC epoch.
        """
        # Double check timestamp, can't imagine why it would be signed
        self._output_buffer.extend(pack(long(timegm(t.timetuple()))))
        return self

    # NOTE: coding to http://dev.rabbitmq.com/wiki/Amqp091Errata#section_3 and
    # NOT spec 0.9.1. It seems that Rabbit and other brokers disagree on this
    # section for now.
    def write_table(self, d):
        """
        Write out a Python dictionary made of up string keys, and values
        that are strings, signed integers, Decimal, datetime.datetime, or
        sub-dictionaries following the same constraints.
        """
        # HACK: encoding of AMQP tables is broken because it requires the
        # length of the /encoded/ data instead of the number of items. To
        # support streaming, fiddle with cursor position, rewinding to write
        # the real length of the data. Generally speaking, I'm not a fan of
        # the AMQP encoding scheme, it could be much faster.
        table_len_pos = len(self._output_buffer)
        self.write_long(0)
        table_data_pos = len(self._output_buffer)

        for key, value in d.iteritems():
            self._write_item(key, value)

        table_end_pos = len(self._output_buffer)
        table_len = table_end_pos - table_data_pos

        self.write_long_at(table_len, table_len_pos)
        return self

    def _write_item(self, key, value):
        self.write_shortstr(key)
        self._write_field(value)

    def _write_field(self, value):
        writer = self.field_type_map.get(type(value))
        if writer:
            writer(self, value)
        else:
            for kls, writer in self.field_type_map.items():
                if isinstance(value, kls):
                    writer(self, value)
                    break
            else:
                # Write a None because we've already written a key
                self._field_none(value)

    def _field_bool(self, val, pack=Struct('B').pack):
        self._output_buffer.append('t')
        self._output_buffer.append(pack(True if val else False))

    def _field_int(self, val, short_pack=Struct('>h').pack,
                   int_pack=Struct('>i').pack, long_pack=Struct('>q').pack):
        if -2 ** 15 <= val < 2 ** 15:
            self._output_buffer.append('s')
            self._output_buffer.extend(short_pack(val))
        elif -2 ** 31 <= val < 2 ** 31:
            self._output_buffer.append('I')
            self._output_buffer.extend(int_pack(val))
        else:
            self._output_buffer.append('l')
            self._output_buffer.extend(long_pack(val))

    def _field_double(self, val, pack=Struct('>d').pack):
        self._output_buffer.append('d')
        self._output_buffer.extend(pack(val))

    # Coding to http://dev.rabbitmq.com/wiki/Amqp091Errata#section_3 which
    # differs from spec in that the value is signed.
    def _field_decimal(self, val, exp_pack=Struct('B').pack,
                       dig_pack=Struct('>i').pack):
        self._output_buffer.append('D')
        sign, digits, exponent = val.as_tuple()
        v = 0
        for d in digits:
            v = (v * 10) + d
        if sign:
            v = -v
        self._output_buffer.append(exp_pack(-exponent))
        self._output_buffer.extend(dig_pack(v))

    def _field_str(self, val):
        self._output_buffer.append('S')
        self.write_longstr(val)

    def _field_unicode(self, val):
        val = val.encode('utf-8')
        self._output_buffer.append('S')
        self.write_longstr(val)

    def _field_timestamp(self, val):
        self._output_buffer.append('T')
        self.write_timestamp(val)

    def _field_table(self, val):
        self._output_buffer.append('F')
        self.write_table(val)

    def _field_none(self, val):
        self._output_buffer.append('V')

    def _field_bytearray(self, val):
        self._output_buffer.append('x')
        self.write_longstr(val)

    def _field_iterable(self, val):
        self._output_buffer.append('A')
        for x in val:
            self._write_field(x)

    field_type_map = {
        bool: _field_bool,
        int: _field_int,
        long: _field_int,
        float: _field_double,
        Decimal: _field_decimal,
        str: _field_str,
        unicode: _field_unicode,
        datetime: _field_timestamp,
        dict: _field_table,
        type(None): _field_none,
        bytearray: _field_bytearray,
    }

    # 0.9.1 spec mapping
    # field_type_map = {
    #   bool      : _field_bool,
    #   int       : _field_int,
    #   long      : _field_int,
    #   float     : _field_double,
    #   Decimal   : _field_decimal,
    #   str       : _field_str,
    #   unicode   : _field_unicode,
    #   datetime  : _field_timestamp,
    #   dict      : _field_table,
    #   None      : _field_none,
    #   bytearray : _field_bytearray,
    #   list      : _field_iterable,
    #   tuple     : _field_iterable,
    #   set       : _field_iterable,
    # }



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.transports.transport import Transport

import errno
import socket


class SocketTransport(Transport):

    '''
    A simple blocking socket transport.
    '''

    def __init__(self, *args):
        super(SocketTransport, self).__init__(*args)
        self._synchronous = True
        self._buffer = bytearray()

    ###
    # Transport API
    ###
    def connect(self, (host, port), klass=socket.socket):
        '''
        Connect assuming a host and port tuple.
        '''
        self._host = "%s:%s" % (host, port)
        self._sock = klass()
        self._sock.setblocking(True)
        self._sock.settimeout(self.connection._connect_timeout)
        if self.connection._sock_opts:
            for k, v in self.connection._sock_opts.iteritems():
                family, type = k
                self._sock.setsockopt(family, type, v)
        self._sock.connect((host, port))

        # After connecting, switch to full-blocking mode.
        self._sock.settimeout(None)

    def read(self, timeout=None):
        '''
        Read from the transport. If timeout>0, will only block for `timeout`
        seconds.
        '''
        e = None
        if not hasattr(self, '_sock'):
            return None

        try:
            # Note that we ignore both None and 0, i.e. we either block with a
            # timeout or block completely and let gevent sort it out.
            if timeout:
                self._sock.settimeout(timeout)
            else:
                self._sock.settimeout(None)
            data = self._sock.recv(
                self._sock.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF))

            if len(data):
                if self.connection.debug > 1:
                    self.connection.logger.debug(
                        'read %d bytes from %s' % (len(data), self._host))
                if len(self._buffer):
                    self._buffer.extend(data)
                    data = self._buffer
                    self._buffer = bytearray()
                return data

            # Note that no data means the socket is closed and we'll mark that
            # below

        except socket.timeout as e:
            # Note that this is implemented differently and though it would be
            # caught as an EnvironmentError, it has no errno. Not sure whose
            # fault that is.
            return None

        except EnvironmentError as e:
            # thrown if we have a timeout and no data
            if e.errno in (errno.EAGAIN, errno.EWOULDBLOCK, errno.EINTR):
                return None

            self.connection.logger.exception(
                'error reading from %s' % (self._host))

        self.connection.transport_closed(
            msg='error reading from %s' % (self._host))
        if e:
            raise

    def buffer(self, data):
        '''
        Buffer unused bytes from the input stream.
        '''
        if not hasattr(self, '_sock'):
            return None

        # data will always be a byte array
        if len(self._buffer):
            self._buffer.extend(data)
        else:
            self._buffer = bytearray(data)

    def write(self, data):
        '''
        Write some bytes to the transport.
        '''
        if not hasattr(self, '_sock'):
            return None

        try:
            self._sock.sendall(data)

            if self.connection.debug > 1:
                self.connection.logger.debug(
                    'sent %d bytes to %s' % (len(data), self._host))

            return
        except EnvironmentError:
            # sockets raise this type of error, and since if sendall() fails
            # we're left in an indeterminate state, assume that any error we
            # catch means that the connection is dead. Note that this
            # assumption requires this to be a blocking socket; if we ever
            # support non-blocking in this class then this whole method has
            # to change a lot.
            self.connection.logger.exception(
                'error writing to %s' % (self._host))

        self.connection.transport_closed(
            msg='error writing to %s' % (self._host))

    def disconnect(self):
        '''
        Disconnect from the transport. Typically socket.close(). This call is
        welcome to raise exceptions, which the Connection will catch.
        '''
        if not hasattr(self, '_sock'):
            return None

        try:
            self._sock.close()
        finally:
            self._sock = None



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''


class ConnectionError(Exception):

    '''Base class for all connection errors.'''


class ConnectionClosed(ConnectionError):

    '''The connection is closed.  Fatal.'''


class ChannelError(Exception):

    '''Base class for all channel errors.'''


class ChannelClosed(ChannelError):

    '''The channel is closed. Fatal.'''



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

from haigha.writer import Writer
from haigha.reader import Reader
from haigha.frames.frame import Frame


class HeaderFrame(Frame):

    '''
    Header frame for content.
    '''
    PROPERTIES = [
        ('content_type', 'shortstr', Reader.read_shortstr,
         Writer.write_shortstr, 1 << 15),
        ('content_encoding', 'shortstr', Reader.read_shortstr,
         Writer.write_shortstr, 1 << 14),
        ('application_headers', 'table',
            Reader.read_table, Writer.write_table, 1 << 13),
        ('delivery_mode', 'octet', Reader.read_octet,
            Writer.write_octet, 1 << 12),
        ('priority', 'octet', Reader.read_octet, Writer.write_octet, 1 << 11),
        ('correlation_id', 'shortstr', Reader.read_shortstr,
            Writer.write_shortstr, 1 << 10),
        ('reply_to', 'shortstr', Reader.read_shortstr,
            Writer.write_shortstr, 1 << 9),
        ('expiration', 'shortstr', Reader.read_shortstr,
            Writer.write_shortstr, 1 << 8),
        ('message_id', 'shortstr', Reader.read_shortstr,
            Writer.write_shortstr, 1 << 7),
        ('timestamp', 'timestamp', Reader.read_timestamp,
            Writer.write_timestamp, 1 << 6),
        ('type', 'shortstr', Reader.read_shortstr,
            Writer.write_shortstr, 1 << 5),
        ('user_id', 'shortstr', Reader.read_shortstr,
            Writer.write_shortstr, 1 << 4),
        ('app_id', 'shortstr', Reader.read_shortstr,
            Writer.write_shortstr, 1 << 3),
        ('cluster_id', 'shortstr', Reader.read_shortstr,
            Writer.write_shortstr, 1 << 2)
    ]
    DEFAULT_PROPERTIES = True

    @classmethod
    def type(cls):
        return 2

    @property
    def class_id(self):
        return self._class_id

    @property
    def weight(self):
        return self._weight

    @property
    def size(self):
        return self._size

    @property
    def properties(self):
        return self._properties

    @classmethod
    def parse(self, channel_id, payload):
        '''
        Parse a header frame for a channel given a Reader payload.
        '''
        class_id = payload.read_short()
        weight = payload.read_short()
        size = payload.read_longlong()
        properties = {}

        # The AMQP spec is overly-complex when it comes to handling header
        # frames. The spec says that in addition to the first 16bit field,
        # additional ones can follow which /may/ then be in the property list
        # (because bit flags aren't in the list). Properly implementing custom
        # values requires the ability change the properties and their types,
        # which someone is welcome to do, but seriously, what's the point?
        # Because the complexity of parsing and writing this frame directly
        # impacts the speed at which messages can be processed, there are two
        # branches for both a fast parse which assumes no changes to the
        # properties and a slow parse. For now it's up to someone using custom
        # headers to flip the flag.
        if self.DEFAULT_PROPERTIES:
            flag_bits = payload.read_short()
            for key, proptype, rfunc, wfunc, mask in self.PROPERTIES:
                if flag_bits & mask:
                    properties[key] = rfunc(payload)
        else:
            flags = []
            while True:
                flag_bits = payload.read_short()
                flags.append(flag_bits)
                if flag_bits & 1 == 0:
                    break

            shift = 0
            for key, proptype, rfunc, wfunc, mask in self.PROPERTIES:
                if shift == 0:
                    if not flags:
                        break
                    flag_bits, flags = flags[0], flags[1:]
                    shift = 15
                if flag_bits & (1 << shift):
                    properties[key] = rfunc(payload)
                shift -= 1

        return HeaderFrame(channel_id, class_id, weight, size, properties)

    def __init__(self, channel_id, class_id, weight, size, properties={}):
        Frame.__init__(self, channel_id)
        self._class_id = class_id
        self._weight = weight
        self._size = size
        self._properties = properties

    def __str__(self):
        return "%s[channel: %d, class_id: %d, weight: %d, size: %d, properties: %s]" % (
            self.__class__.__name__, self.channel_id, self._class_id,
            self._weight, self._size, self._properties)

    def write_frame(self, buf):
        '''
        Write the frame into an existing buffer.
        '''
        writer = Writer(buf)
        writer.write_octet(self.type())
        writer.write_short(self.channel_id)

        # Track the position where we're going to write the total length
        # of the frame arguments.
        stream_args_len_pos = len(buf)
        writer.write_long(0)

        stream_method_pos = len(buf)

        writer.write_short(self._class_id)
        writer.write_short(self._weight)
        writer.write_longlong(self._size)

        # Like frame parsing, branch to faster code for default properties
        if self.DEFAULT_PROPERTIES:
            # Track the position where we're going to write the flags.
            flags_pos = len(buf)
            writer.write_short(0)
            flag_bits = 0
            for key, proptype, rfunc, wfunc, mask in self.PROPERTIES:
                val = self._properties.get(key, None)
                if val is not None:
                    flag_bits |= mask
                    wfunc(writer, val)
            writer.write_short_at(flag_bits, flags_pos)
        else:
            shift = 15
            flag_bits = 0
            flags = []
            stack = deque()
            for key, proptype, rfunc, wfunc, mask in self.PROPERTIES:
                val = self._properties.get(key, None)
                if val is not None:
                    if shift == 0:
                        flags.append(flag_bits)
                        flag_bits = 0
                        shift = 15

                    flag_bits |= (1 << shift)
                    stack.append((wfunc, val))

                shift -= 1

            flags.append(flag_bits)
            for flag_bits in flags:
                writer.write_short(flag_bits)
            for method, val in stack:
                method(writer, val)

        # Write the total length back at the beginning of the frame
        stream_len = len(buf) - stream_method_pos
        writer.write_long_at(stream_len, stream_args_len_pos)

        writer.write_octet(0xce)

HeaderFrame.register()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''


class Message(object):

    '''
    Represents an AMQP message.
    '''

    def __init__(self, body='', delivery_info=None, **properties):
        if isinstance(body, unicode):
            if 'content_encoding' not in properties:
                properties['content_encoding'] = 'utf-8'
            body = body.encode(properties['content_encoding'])

        if not isinstance(body, (str, unicode, bytearray)):
            raise TypeError("Invalid message content type %s" % (type(body)))

        self._body = body
        self._delivery_info = delivery_info
        self._properties = properties

    @property
    def body(self):
        return self._body

    def __len__(self):
        return len(self._body)

    def __nonzero__(self):
        '''Have to define this because length is defined.'''
        return True

    def __eq__(self, other):
        if isinstance(other, Message):
            return self._properties == other._properties and \
                self._body == other._body
        return False

    @property
    def delivery_info(self):
        return self._delivery_info

    @property
    def properties(self):
        return self._properties

    def __str__(self):
        return "Message[body: %s, delivery_info: %s, properties: %s]" %\
            (str(self._body).encode('string_escape'),
             self._delivery_info, self._properties)



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''


class Transport(object):

    '''
    Base class and API for Transports
    '''

    def __init__(self, connection):
        '''
        Initialize a transport on a haigha.Connection instance.
        '''
        self._connection = connection

    @property
    def synchronous(self):
        '''Return True if this is a synchronous transport, False otherwise.'''
        # Note that subclasses must define this.
        return self._synchronous

    @property
    def connection(self):
        return self._connection

    def process_channels(self, channels):
        '''
        Process a set of channels by calling Channel.process_frames() on each.
        Some transports may choose to do this in unique ways, such as through
        a pool of threads.

        The default implementation will simply iterate over them and call
        process_frames() on each.
        '''
        for channel in channels:
            channel.process_frames()

    def read(self, timeout=None):
        '''
        Read from the transport. If no data is available, should return None.
        The return value can be any data type that is supported by the
        haigha.Reader class.

        Caller passes in an optional timeout. Each transport determines how to
        implement this.
        '''
        return None

    def buffer(self, data):
        '''
        Buffer unused bytes from the input stream.
        '''

    def write(self, data):
        '''
        Write some bytes to the transport.
        '''

    def disconnect(self):
        '''
        Disconnect from the transport. Typically socket.close(). This call is
        welcome to raise exceptions, which the Connection will catch.

        The transport is encouraged to allow for any pending writes to complete
        before closing the socket.
        '''



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

from haigha.connection import Connection
from haigha.classes.basic_class import BasicClass
from haigha.classes.exchange_class import ExchangeClass
from haigha.classes.protocol_class import ProtocolClass
from haigha.writer import Writer
from haigha.frames.method_frame import MethodFrame


class RabbitConnection(Connection):

    '''
    A connection specific to RabbitMQ that supports its extensions.
    '''

    def __init__(self, **kwargs):
        '''
        Initialize the connection
        '''
        class_map = kwargs.get('class_map', {}).copy()
        class_map.setdefault(40, RabbitExchangeClass)
        class_map.setdefault(60, RabbitBasicClass)
        class_map.setdefault(85, RabbitConfirmClass)
        kwargs['class_map'] = class_map

        super(RabbitConnection, self).__init__(**kwargs)


class RabbitExchangeClass(ExchangeClass):

    '''
    Exchange class Rabbit extensions
    '''

    def __init__(self, *args, **kwargs):
        super(RabbitExchangeClass, self).__init__(*args, **kwargs)
        self.dispatch_map[31] = self._recv_bind_ok
        self.dispatch_map[51] = self._recv_unbind_ok

        self._bind_cb = deque()
        self._unbind_cb = deque()

    # I hate the code copying here. Probably a better solution, like
    # functools.
    def declare(self, exchange, type, passive=False, durable=False,
                auto_delete=True, internal=False, nowait=True,
                arguments=None, ticket=None, cb=None):
        """
        Declare the exchange.

        exchange - The name of the exchange to declare
        type - One of
        """
        nowait = nowait and self.allow_nowait() and not cb

        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(exchange).\
            write_shortstr(type).\
            write_bits(passive, durable, auto_delete, internal, nowait).\
            write_table(arguments or {})
        self.send_frame(MethodFrame(self.channel_id, 40, 10, args))

        if not nowait:
            self._declare_cb.append(cb)
            self.channel.add_synchronous_cb(self._recv_declare_ok)

    def bind(self, exchange, source, routing_key='', nowait=True,
             arguments={}, ticket=None, cb=None):
        '''
        Bind an exchange to another.
        '''
        nowait = nowait and self.allow_nowait() and not cb

        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(exchange).\
            write_shortstr(source).\
            write_shortstr(routing_key).\
            write_bit(nowait).\
            write_table(arguments or {})
        self.send_frame(MethodFrame(self.channel_id, 40, 30, args))

        if not nowait:
            self._bind_cb.append(cb)
            self.channel.add_synchronous_cb(self._recv_bind_ok)

    def _recv_bind_ok(self, _method_frame):
        '''Confirm exchange bind.'''
        cb = self._bind_cb.popleft()
        if cb:
            cb()

    def unbind(self, exchange, source, routing_key='', nowait=True,
               arguments={}, ticket=None, cb=None):
        '''
        Unbind an exchange from another.
        '''
        nowait = nowait and self.allow_nowait() and not cb

        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(exchange).\
            write_shortstr(source).\
            write_shortstr(routing_key).\
            write_bit(nowait).\
            write_table(arguments or {})
        self.send_frame(MethodFrame(self.channel_id, 40, 40, args))

        if not nowait:
            self._unbind_cb.append(cb)
            self.channel.add_synchronous_cb(self._recv_unbind_ok)

    def _recv_unbind_ok(self, _method_frame):
        '''Confirm exchange unbind.'''
        cb = self._unbind_cb.popleft()
        if cb:
            cb()


class RabbitBasicClass(BasicClass):

    '''
    Support Rabbit extensions to Basic class.
    '''

    def __init__(self, *args, **kwargs):
        super(RabbitBasicClass, self).__init__(*args, **kwargs)
        self.dispatch_map[80] = self._recv_ack
        self.dispatch_map[120] = self._recv_nack

        self._ack_listener = None
        self._nack_listener = None

        self._msg_id = 0
        self._last_ack_id = 0

    def set_ack_listener(self, cb):
        '''
        Set a callback for ack listening, to be used when the channel is
        in publisher confirm mode. Will be called with a single integer
        argument which is the id of the message as returned from publish().

        cb(message_id)
        '''
        self._ack_listener = cb

    def set_nack_listener(self, cb):
        '''
        Set a callbnack for nack listening, to be used when the channel is
        in publisher confirm mode. Will be called with a single integer
        argument which is the id of the message as returned from publish()
        and a boolean flag indicating if it can be requeued.

        cb(message_id, reque)
        '''
        self._nack_listener = cb

    # Probably a better solution here, like functools
    def publish(self, *args, **kwargs):
        '''
        Publish a message. Will return the id of the message if publisher
        confirmations are enabled, else will return 0.
        '''
        if self.channel.confirm._enabled:
            self._msg_id += 1
        super(RabbitBasicClass, self).publish(*args, **kwargs)
        return self._msg_id

    def _recv_ack(self, method_frame):
        '''Receive an ack from the broker.'''
        if self._ack_listener:
            delivery_tag = method_frame.args.read_longlong()
            multiple = method_frame.args.read_bit()
            if multiple:
                while self._last_ack_id < delivery_tag:
                    self._last_ack_id += 1
                    self._ack_listener(self._last_ack_id)
            else:
                self._last_ack_id = delivery_tag
                self._ack_listener(self._last_ack_id)

    def nack(self, delivery_tag, multiple=False, requeue=False):
        '''Send a nack to the broker.'''
        args = Writer()
        args.write_longlong(delivery_tag).\
            write_bits(multiple, requeue)

        self.send_frame(MethodFrame(self.channel_id, 60, 120, args))

    def _recv_nack(self, method_frame):
        '''Receive a nack from the broker.'''
        if self._nack_listener:
            delivery_tag = method_frame.args.read_longlong()
            multiple, requeue = method_frame.args.read_bits(2)
            if multiple:
                while self._last_ack_id < delivery_tag:
                    self._last_ack_id += 1
                    self._nack_listener(self._last_ack_id, requeue)
            else:
                self._last_ack_id = delivery_tag
                self._nack_listener(self._last_ack_id, requeue)


class RabbitConfirmClass(ProtocolClass):

    '''
    Implementation of Rabbit's confirm class.
    '''

    def __init__(self, *args, **kwargs):
        super(RabbitConfirmClass, self).__init__(*args, **kwargs)
        self.dispatch_map = {
            11: self._recv_select_ok,
        }

        self._enabled = False
        self._select_cb = deque()

    @property
    def name(self):
        return 'confirm'

    def select(self, nowait=True, cb=None):
        '''
        Set this channel to use publisher confirmations.
        '''
        nowait = nowait and self.allow_nowait() and not cb

        if not self._enabled:
            self._enabled = True
            self.channel.basic._msg_id = 0
            self.channel.basic._last_ack_id = 0
            args = Writer()
            args.write_bit(nowait)

            self.send_frame(MethodFrame(self.channel_id, 85, 10, args))

            if not nowait:
                self._select_cb.append(cb)
                self.channel.add_synchronous_cb(self._recv_select_ok)

    def _recv_select_ok(self, _method_frame):
        cb = self._select_cb.popleft()
        if cb:
            cb()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

import warnings

from haigha.transports.socket_transport import SocketTransport

try:
    import gevent
    from gevent.event import Event
    try:
        # Semaphore moved here since gevent-1.0b2
        from gevent.lock import Semaphore
    except ImportError:
        from gevent.coros import Semaphore
    from gevent import socket
    from gevent import pool
except ImportError:
    warnings.warn('Failed to load gevent modules')
    gevent = None
    Event = None
    Semaphore = None
    socket = None
    pool = None


class GeventTransport(SocketTransport):

    '''
    Transport using gevent backend. It relies on gevent's implementation of
    sendall to send whole frames at a time. On the input side, it uses a gevent
    semaphore to ensure exclusive access to the socket and input buffer.
    '''

    def __init__(self, *args, **kwargs):
        super(GeventTransport, self).__init__(*args)

        self._synchronous = False
        self._read_lock = Semaphore()
        self._write_lock = Semaphore()
        self._read_wait = Event()

    ###
    # Transport API
    ###

    def connect(self, (host, port)):
        '''
        Connect using a host,port tuple
        '''
        super(GeventTransport, self).connect((host, port), klass=socket.socket)

    def read(self, timeout=None):
        '''
        Read from the transport. If no data is available, should return None.
        If timeout>0, will only block for `timeout` seconds.
        '''
        # If currently locked, another greenlet is trying to read, so yield
        # control and then return none. Required if a Connection is configured
        # to be synchronous, a sync callback is trying to read, and there's
        # another read loop running read_frames. Without it, the run loop will
        # release the lock but then immediately acquire it again. Yielding
        # control in the reading thread after bytes are read won't fix
        # anything, because it's quite possible the bytes read resulted in a
        # frame that satisfied the synchronous callback, and so this needs to
        # return immediately to first check the current status of synchronous
        # callbacks before attempting to read again.
        if self._read_lock.locked():
            self._read_wait.wait(timeout)
            return None

        self._read_lock.acquire()
        try:
            return super(GeventTransport, self).read(timeout=timeout)
        finally:
            self._read_lock.release()
            self._read_wait.set()
            self._read_wait.clear()

    def buffer(self, data):
        '''
        Buffer unused bytes from the input stream.
        '''
        self._read_lock.acquire()
        try:
            return super(GeventTransport, self).buffer(data)
        finally:
            self._read_lock.release()

    def write(self, data):
        '''
        Write some bytes to the transport.
        '''
        # MUST use a lock here else gevent could raise an exception if 2
        # greenlets try to write at the same time. I was hoping that
        # sendall() would do that blocking for me, but I guess not. May
        # require an eventsocket-like buffer to speed up under high load.
        self._write_lock.acquire()
        try:
            return super(GeventTransport, self).write(data)
        finally:
            self._write_lock.release()


class GeventPoolTransport(GeventTransport):

    def __init__(self, *args, **kwargs):
        super(GeventPoolTransport, self).__init__(*args)

        self._pool = kwargs.get('pool', None)
        if not self._pool:
            self._pool = gevent.pool.Pool()

    @property
    def pool(self):
        '''Get a handle to the gevent pool.'''
        return self._pool

    def process_channels(self, channels):
        '''
        Process a set of channels by calling Channel.process_frames() on each.
        Some transports may choose to do this in unique ways, such as through
        a pool of threads.

        The default implementation will simply iterate over them and call
        process_frames() on each.
        '''
        for channel in channels:
            self._pool.spawn(channel.process_frames)



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

import struct
import sys
from collections import deque
from haigha.reader import Reader


class Frame(object):

    '''
    Base class for a frame.
    '''

    # Exceptions
    class FrameError(Exception):

        '''Base class for all frame errors'''
    class FormatError(FrameError):

        '''The frame was mal-formed.'''
    class InvalidFrameType(FrameError):

        '''The frame type is unknown.'''

    # Class data
    _frame_type_map = {}

    # Class methods
    @classmethod
    def register(cls):
        '''
        Register a frame type.
        '''
        cls._frame_type_map[cls.type()] = cls

    @classmethod
    def type(self):
        '''
        Fetch the type of this frame.  Should be an octet.
        '''
        raise NotImplementedError()

    @classmethod
    def read_frames(cls, reader):
        '''
        Read one or more frames from an IO stream.  Buffer must support file
        object interface.

        After reading, caller will need to check if there are bytes remaining
        in the stream. If there are, then that implies that there is one or
        more incomplete frames and more data needs to be read.  The position
        of the cursor in the frame stream will mark the point at which the
        last good frame was read. If the caller is expecting a sequence of
        frames and only received a part of that sequence, they are responsible
        for buffering those frames until the rest of the frames in the sequence
        have arrived.
        '''
        rval = deque()

        while True:
            frame_start_pos = reader.tell()
            try:
                frame = Frame._read_frame(reader)
            except Reader.BufferUnderflow:
                # No more data in the stream
                frame = None
            except Reader.ReaderError as e:
                # Some other format error
                raise Frame.FormatError, str(e), sys.exc_info()[-1]
            except struct.error as e:
                raise Frame.FormatError, str(e), sys.exc_info()[-1]

            if frame is None:
                reader.seek(frame_start_pos)
                break

            rval.append(frame)

        return rval

    @classmethod
    def _read_frame(cls, reader):
        '''
        Read a single frame from a Reader.  Will return None if there is an
        incomplete frame in the stream.

        Raise MissingFooter if there's a problem reading the footer byte.
        '''
        frame_type = reader.read_octet()
        channel_id = reader.read_short()
        size = reader.read_long()

        payload = Reader(reader, reader.tell(), size)

        # Seek to end of payload
        reader.seek(size, 1)

        ch = reader.read_octet()  # footer
        if ch != 0xce:
            raise Frame.FormatError(
                'Framing error, unexpected byte: %x.  frame type %x. channel %d, payload size %d',
                ch, frame_type, channel_id, size)

        frame_class = cls._frame_type_map.get(frame_type)
        if not frame_class:
            raise Frame.InvalidFrameType("Unknown frame type %x", frame_type)
        return frame_class.parse(channel_id, payload)

    # Instance methods
    def __init__(self, channel_id=-1):
        self._channel_id = channel_id

    @classmethod
    def parse(cls, channel_id, payload):
        '''
        Subclasses need to implement parsing of their frames.  Should return
        a new instance of their type.
        '''
        raise NotImplementedError()

    @property
    def channel_id(self):
        return self._channel_id

    def __str__(self):
        return "%s[channel: %d]" % (self.__class__.__name__, self.channel_id)

    def __repr__(self):
        # Have to actually call the method rather than __repr__==__str__
        # because subclasses overload __str__
        return str(self)

    def write_frame(self, stream):
        '''
        Write this frame.
        '''
        raise NotImplementedError()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.writer import Writer
from haigha.frames.method_frame import MethodFrame
from haigha.classes.protocol_class import ProtocolClass

from collections import deque


class QueueClass(ProtocolClass):

    '''
    Implements the AMQP Queue class
    '''

    def __init__(self, *args, **kwargs):
        super(QueueClass, self).__init__(*args, **kwargs)
        self.dispatch_map = {
            11: self._recv_declare_ok,
            21: self._recv_bind_ok,
            31: self._recv_purge_ok,
            41: self._recv_delete_ok,
            51: self._recv_unbind_ok,
        }

        self._declare_cb = deque()
        self._bind_cb = deque()
        self._unbind_cb = deque()
        self._delete_cb = deque()
        self._purge_cb = deque()

    @property
    def name(self):
        return 'queue'

    def _cleanup(self):
        '''
        Cleanup all the local data.
        '''
        self._declare_cb = None
        self._bind_cb = None
        self._unbind_cb = None
        self._delete_cb = None
        self._purge_cb = None
        super(QueueClass, self)._cleanup()

    def declare(self, queue='', passive=False, durable=False,
                exclusive=False, auto_delete=True, nowait=True,
                arguments={}, ticket=None, cb=None):
        '''
        Declare a queue. By default is asynchronoous but will be synchronous
        if nowait=False or a callback is defined. In synchronous mode,
        returns (message_count, consumer_count)

        queue - The name of the queue
        cb - An optional method which will be called with
              (queue_name, msg_count, consumer_count) if nowait=False
        '''
        nowait = nowait and self.allow_nowait() and not cb

        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(queue).\
            write_bits(passive, durable, exclusive, auto_delete, nowait).\
            write_table(arguments)
        self.send_frame(MethodFrame(self.channel_id, 50, 10, args))

        if not nowait:
            self._declare_cb.append(cb)
            return self.channel.add_synchronous_cb(self._recv_declare_ok)

    def _recv_declare_ok(self, method_frame):
        queue = method_frame.args.read_shortstr()
        message_count = method_frame.args.read_long()
        consumer_count = method_frame.args.read_long()

        cb = self._declare_cb.popleft()
        if cb:
            cb(queue, message_count, consumer_count)
        return queue, message_count, consumer_count

    def bind(self, queue, exchange, routing_key='', nowait=True, arguments={},
             ticket=None, cb=None):
        '''
        bind to a queue.
        '''
        nowait = nowait and self.allow_nowait() and not cb

        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(queue).\
            write_shortstr(exchange).\
            write_shortstr(routing_key).\
            write_bit(nowait).\
            write_table(arguments)
        self.send_frame(MethodFrame(self.channel_id, 50, 20, args))

        if not nowait:
            self._bind_cb.append(cb)
            self.channel.add_synchronous_cb(self._recv_bind_ok)

    def _recv_bind_ok(self, _method_frame):
        # No arguments defined.
        cb = self._bind_cb.popleft()
        if cb:
            cb()

    def unbind(self, queue, exchange, routing_key='', arguments={},
               ticket=None, cb=None):
        '''
        Unbind a queue from an exchange.  This is always synchronous.
        '''
        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(queue).\
            write_shortstr(exchange).\
            write_shortstr(routing_key).\
            write_table(arguments)
        self.send_frame(MethodFrame(self.channel_id, 50, 50, args))

        self._unbind_cb.append(cb)
        self.channel.add_synchronous_cb(self._recv_unbind_ok)

    def _recv_unbind_ok(self, _method_frame):
        # No arguments defined
        cb = self._unbind_cb.popleft()
        if cb:
            cb()

    def purge(self, queue, nowait=True, ticket=None, cb=None):
        '''
        Purge all messages in a queue.
        '''
        nowait = nowait and self.allow_nowait() and not cb

        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(queue).\
            write_bit(nowait)
        self.send_frame(MethodFrame(self.channel_id, 50, 30, args))

        if not nowait:
            self._purge_cb.append(cb)
            return self.channel.add_synchronous_cb(self._recv_purge_ok)

    def _recv_purge_ok(self, method_frame):
        message_count = method_frame.args.read_long()
        cb = self._purge_cb.popleft()
        if cb:
            cb(message_count)
        return message_count

    def delete(self, queue, if_unused=False, if_empty=False, nowait=True,
               ticket=None, cb=None):
        '''
        queue delete.
        '''
        nowait = nowait and self.allow_nowait() and not cb

        args = Writer()
        args.write_short(ticket or self.default_ticket).\
            write_shortstr(queue).\
            write_bits(if_unused, if_empty, nowait)
        self.send_frame(MethodFrame(self.channel_id, 50, 40, args))

        if not nowait:
            self._delete_cb.append(cb)
            return self.channel.add_synchronous_cb(self._recv_delete_ok)

    def _recv_delete_ok(self, method_frame):
        message_count = method_frame.args.read_long()
        cb = self._delete_cb.popleft()
        if cb:
            cb(message_count)
        return message_count



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.channel import Channel
from haigha.frames.frame import Frame
from haigha.frames.heartbeat_frame import HeartbeatFrame
from haigha.frames.method_frame import MethodFrame
from haigha.classes.basic_class import BasicClass
from haigha.classes.channel_class import ChannelClass
from haigha.classes.exchange_class import ExchangeClass
from haigha.classes.queue_class import QueueClass
from haigha.classes.transaction_class import TransactionClass
from haigha.writer import Writer
from haigha.reader import Reader
from haigha.transports.transport import Transport
from exceptions import ConnectionError, ConnectionClosed

import haigha
import time

from logging import root as root_logger

# From the rabbitmq mailing list
# AMQP1180 = 0-8
# AMQP1109 = 0-9
# AMQP0091 = 0-9-1
# http://lists.rabbitmq.com/pipermail/rabbitmq-discuss/2010-July/008231.html
# PROTOCOL_HEADER = 'AMQP\x01\x01\x09\x01'
PROTOCOL_HEADER = 'AMQP\x00\x00\x09\x01'

# Client property info that gets sent to the server on connection startup
LIBRARY_PROPERTIES = {
    'library': 'Haigha',
    'library_version': haigha.__version__,
}


class Connection(object):

    class TooManyChannels(ConnectionError):
        '''This connection has too many channels open.  Non-fatal.'''

    class InvalidChannel(ConnectionError):
        '''
        The channel id does not correspond to an existing channel.  Non-fatal.
        '''

    def __init__(self, **kwargs):
        '''
        Initialize the connection.
        '''
        self._debug = kwargs.get('debug', False)
        self._logger = kwargs.get('logger', root_logger)

        self._user = kwargs.get('user', 'guest')
        self._password = kwargs.get('password', 'guest')
        self._host = kwargs.get('host', 'localhost')
        self._port = kwargs.get('port', 5672)
        self._vhost = kwargs.get('vhost', '/')

        self._connect_timeout = kwargs.get('connect_timeout', 5)
        self._sock_opts = kwargs.get('sock_opts')
        self._sock = None
        self._heartbeat = kwargs.get('heartbeat')
        self._open_cb = kwargs.get('open_cb')
        self._close_cb = kwargs.get('close_cb')

        self._login_method = kwargs.get('login_method', 'AMQPLAIN')
        self._locale = kwargs.get('locale', 'en_US')
        self._client_properties = kwargs.get('client_properties')

        self._properties = LIBRARY_PROPERTIES.copy()
        if self._client_properties:
            self._properties.update(self._client_properties)

        self._closed = False
        self._connected = False
        self._close_info = {
            'reply_code': 0,
            'reply_text': 'first connect',
            'class_id': 0,
            'method_id': 0
        }

        # Not sure what's better here, setdefaults or require the caller to
        # pass the whole thing in.
        self._class_map = kwargs.get('class_map', {}).copy()
        self._class_map.setdefault(20, ChannelClass)
        self._class_map.setdefault(40, ExchangeClass)
        self._class_map.setdefault(50, QueueClass)
        self._class_map.setdefault(60, BasicClass)
        self._class_map.setdefault(90, TransactionClass)

        self._channels = {
            0: ConnectionChannel(self, 0, {})
        }

        # Login response seems a total hack of protocol
        # Skip the length at the beginning
        login_response = Writer()
        login_response.write_table(
            {'LOGIN': self._user, 'PASSWORD': self._password})
        self._login_response = login_response.buffer()[4:]

        self._channel_counter = 0
        self._channel_max = 65535
        self._frame_max = 65535

        self._frames_read = 0
        self._frames_written = 0

        # Default to the socket strategy
        transport = kwargs.get('transport', 'socket')
        if not isinstance(transport, Transport):
            if transport == 'event':
                from haigha.transports.event_transport import EventTransport
                self._transport = EventTransport(self)
            elif transport == 'gevent':
                from haigha.transports.gevent_transport import GeventTransport
                self._transport = GeventTransport(self)
            elif transport == 'gevent_pool':
                from haigha.transports.gevent_transport import \
                    GeventPoolTransport
                self._transport = GeventPoolTransport(self, **kwargs)
            elif transport == 'socket':
                from haigha.transports.socket_transport import SocketTransport
                self._transport = SocketTransport(self)
        else:
            self._transport = transport

        # Set these after the transport is initialized, so that we can access
        # the synchronous property
        self._synchronous = kwargs.get('synchronous', False)
        self._synchronous_connect = kwargs.get(
            'synchronous_connect', False) or self.synchronous

        self._output_frame_buffer = []
        self.connect(self._host, self._port)

    @property
    def logger(self):
        return self._logger

    @property
    def debug(self):
        return self._debug

    @property
    def frame_max(self):
        return self._frame_max

    @property
    def channel_max(self):
        return self._channel_max

    @property
    def frames_read(self):
        '''Number of frames read in the lifetime of this connection.'''
        return self._frames_read

    @property
    def frames_written(self):
        '''Number of frames written in the lifetime of this connection.'''
        return self._frames_written

    @property
    def closed(self):
        '''Return the closed state of the connection.'''
        return self._closed

    @property
    def close_info(self):
        '''
        Return dict with information on why this connection is closed.  Will
        return None if the connections is open.
        '''
        return self._close_info if (
            self._closed or not self._connected) else None

    @property
    def transport(self):
        '''Get the value of the current transport.'''
        return self._transport

    @property
    def synchronous(self):
        '''
        True if transport is synchronous or the connection has been forced
        into synchronous mode, False otherwise.
        '''
        if self._transport is None:
            if self._close_info and len(self._close_info['reply_text']) > 0:
                raise ConnectionClosed("connection is closed: %s : %s" %
                                       (self._close_info['reply_code'],
                                        self._close_info['reply_text']))
            raise ConnectionClosed("connection is closed")
        return self.transport.synchronous or self._synchronous

    def connect(self, host, port):
        '''
        Connect to a host and port.
        '''
        # Clear the connect state immediately since we're no longer connected
        # at this point.
        self._connected = False

        # Only after the socket has connected do we clear this state; closed
        # must be False so that writes can be buffered in writePacket(). The
        # closed state might have been set to True due to a socket error or a
        # redirect.
        self._host = "%s:%d" % (host, port)
        self._closed = False
        self._close_info = {
            'reply_code': 0,
            'reply_text': 'failed to connect to %s' % (self._host),
            'class_id': 0,
            'method_id': 0
        }

        self._transport.connect((host, port))
        self._transport.write(PROTOCOL_HEADER)

        if self._synchronous_connect:
            # Have to queue this callback just after connect, it can't go
            # into the constructor because the channel needs to be
            # "always there" for frame processing, but the synchronous
            # callback can't be added until after the protocol header has
            # been written. This SHOULD be registered before the protocol
            # header is written, in the case where the header bytes are
            # written, but this thread/greenlet/context does not return until
            # after another thread/greenlet/context has read and processed the
            # recv_start frame. Without more re-write to add_sync_cb though,
            # it will block on reading responses that will never arrive
            # because the protocol header isn't written yet. TBD if needs
            # refactoring. Could encapsulate entirely here, wherein
            # read_frames exits if protocol header not yet written. Like other
            # synchronous behaviors, adding this callback will result in a
            # blocking frame read and process loop until _recv_start and any
            # subsequent synchronous callbacks have been processed. In the
            # event that this is /not/ a synchronous transport, but the
            # caller wants the connect to be synchronous so as to ensure that
            # the connection is ready, then do a read frame loop here.
            self._channels[0].add_synchronous_cb(self._channels[0]._recv_start)
            while not self._connected:
                self.read_frames()

    def disconnect(self):
        '''
        Disconnect from the current host, but do not update the closed state.
        After the transport is disconnected, the closed state will be True if
        this is called after a protocol shutdown, or False if the disconnect
        was in error.

        TODO: do we really need closed vs. connected states? this only adds
        complication and the whole reconnect process has been scrapped anyway.

        '''
        self._connected = False
        if self._transport is not None:
            try:
                self._transport.disconnect()
            except Exception:
                self.logger.error(
                    "Failed to disconnect from %s", self._host, exc_info=True)
                raise
            finally:
                self._transport = None

    ###
    # Transport methods
    ###
    def transport_closed(self, **kwargs):
        """
        Called by Transports when they close unexpectedly, not as a result of
        Connection.disconnect().

        TODO: document args
        """
        msg = 'unknown cause'
        self.logger.warning('transport to %s closed : %s' %
                            (self._host, kwargs.get('msg', msg)))
        self._close_info = {
            'reply_code': kwargs.get('reply_code', 0),
            'reply_text': kwargs.get('msg', msg),
            'class_id': kwargs.get('class_id', 0),
            'method_id': kwargs.get('method_id', 0)
        }

        # We're not connected any more, but we're not closed without an
        # explicit close call.
        self._connected = False
        self._transport = None

        # Call back to a user-provided close function
        self._callback_close()

    ###
    # Connection methods
    ###
    def _next_channel_id(self):
        '''Return the next possible channel id.  Is a circular enumeration.'''
        self._channel_counter += 1
        if self._channel_counter >= self._channel_max:
            self._channel_counter = 1
        return self._channel_counter

    def channel(self, channel_id=None, synchronous=False):
        """
        Fetch a Channel object identified by the numeric channel_id, or
        create that object if it doesn't already exist.  If channel_id is not
        None but no channel exists for that id, will raise InvalidChannel.  If
        there are already too many channels open, will raise TooManyChannels.

        If synchronous=True, then the channel will act synchronous in all cases
        where a protocol method supports `nowait=False`, or where there is an
        implied callback in the protocol.
        """
        if channel_id is None:
            # adjust for channel 0
            if len(self._channels) - 1 >= self._channel_max:
                raise Connection.TooManyChannels(
                    "%d channels already open, max %d",
                    len(self._channels) - 1,
                    self._channel_max)
            channel_id = self._next_channel_id()
            while channel_id in self._channels:
                channel_id = self._next_channel_id()
        elif channel_id in self._channels:
            return self._channels[channel_id]
        else:
            raise Connection.InvalidChannel(
                "%s is not a valid channel id", channel_id)

        # Call open() here so that ConnectionChannel doesn't have it called.
        # Could also solve this other ways, but it's a HACK regardless.
        rval = Channel(
            self, channel_id, self._class_map, synchronous=synchronous)
        self._channels[channel_id] = rval
        rval.add_close_listener(self._channel_closed)
        rval.open()
        return rval

    def _channel_closed(self, channel):
        '''
        Close listener on a channel.
        '''
        try:
            del self._channels[channel.channel_id]
        except KeyError:
            pass

    def close(self, reply_code=0, reply_text='', class_id=0, method_id=0,
              disconnect=False):
        '''
        Close this connection.
        '''
        self._close_info = {
            'reply_code': reply_code,
            'reply_text': reply_text,
            'class_id': class_id,
            'method_id': method_id
        }
        if disconnect:
            self._closed = True
            self.disconnect()
            self._callback_close()
        else:
            self._channels[0].close()

    def _callback_open(self):
        '''
        Callback to any open handler that was provided in the ctor. Handler is
        responsible for exceptions.
        '''
        if self._open_cb:
            self._open_cb()

    def _callback_close(self):
        '''
        Callback to any close handler that was provided in the ctor. Handler is
        responsible for exceptions.
        '''
        if self._close_cb:
            self._close_cb()

    def read_frames(self):
        '''
        Read frames from the transport and process them. Some transports may
        choose to do this in the background, in several threads, and so on.
        '''
        # It's possible in a concurrent environment that our transport handle
        # has gone away, so handle that cleanly.
        # TODO: Consider moving this block into Translator base class. In many
        # ways it belongs there. One of the problems though is that this is
        # essentially the read loop. Each Transport has different rules for
        # how to kick this off, and in the case of gevent, this is how a
        # blocking call to read from the socket is kicked off.
        if self._transport is None:
            return

        # Send a heartbeat (if needed)
        self._channels[0].send_heartbeat()

        data = self._transport.read(self._heartbeat)
        if data is None:
            return

        reader = Reader(data)
        p_channels = set()

        try:
            for frame in Frame.read_frames(reader):
                if self._debug > 1:
                    self.logger.debug("READ: %s", frame)
                self._frames_read += 1
                ch = self.channel(frame.channel_id)
                ch.buffer_frame(frame)
                p_channels.add(ch)
        except Frame.FrameError as e:
            # Frame error in the peer, disconnect
            self.close(reply_code=501,
                       reply_text='frame error from %s : %s' % (
                           self._host, str(e)),
                       class_id=0, method_id=0, disconnect=True)
            raise ConnectionClosed("connection is closed: %s : %s" %
                                   (self._close_info['reply_code'],
                                    self._close_info['reply_text']))

        self._transport.process_channels(p_channels)

        # HACK: read the buffer contents and re-buffer.  Would prefer to pass
        # buffer back, but there's no good way of asking the total size of the
        # buffer, comparing to tell(), and then re-buffering.  There's also no
        # ability to clear the buffer up to the current position. It would be
        # awesome if we could free that memory without a new allocation.
        if reader.tell() < len(data):
            self._transport.buffer(data[reader.tell():])

    def _flush_buffered_frames(self):
        '''
        Callback when protocol has been initialized on channel 0 and we're
        ready to send out frames to set up any channels that have been
        created.
        '''
        # In the rare case (a bug) where this is called but send_frame thinks
        # they should be buffered, don't clobber.
        frames = self._output_frame_buffer
        self._output_frame_buffer = []
        for frame in frames:
            self.send_frame(frame)

    def send_frame(self, frame):
        '''
        Send a single frame. If there is no transport or we're not connected
        yet, append to the output buffer, else send immediately to the socket.
        This is called from within the MethodFrames.
        '''
        if self._closed:
            if self._close_info and len(self._close_info['reply_text']) > 0:
                raise ConnectionClosed("connection is closed: %s : %s" %
                                       (self._close_info['reply_code'],
                                        self._close_info['reply_text']))
            raise ConnectionClosed("connection is closed")

        if self._transport is None or \
                (not self._connected and frame.channel_id != 0):
            self._output_frame_buffer.append(frame)
            return

        if self._debug > 1:
            self.logger.debug("WRITE: %s", frame)

        buf = bytearray()
        frame.write_frame(buf)
        if len(buf) > self._frame_max:
            self.close(
                reply_code=501,
                reply_text='attempted to send frame of %d bytes, frame max %d' % (
                    len(buf), self._frame_max),
                class_id=0, method_id=0, disconnect=True)
            raise ConnectionClosed(
                "connection is closed: %s : %s" %
                (self._close_info['reply_code'],
                 self._close_info['reply_text']))
        self._transport.write(buf)

        self._frames_written += 1


class ConnectionChannel(Channel):

    '''
    A special channel for the Connection class.  It's used for performing the
    special methods only available on the main connection channel.  It's also
    partly used to hide the 'connection' protocol implementation, which would
    show up as a property, from the more useful 'connection' property that is
    a handle to a Channel's Connection object.
    '''

    def __init__(self, *args):
        super(ConnectionChannel, self).__init__(*args)

        self._method_map = {
            10: self._recv_start,
            20: self._recv_secure,
            30: self._recv_tune,
            41: self._recv_open_ok,
            50: self._recv_close,
            51: self._recv_close_ok,
        }
        self._last_heartbeat_send = 0

    def dispatch(self, frame):
        '''
        Override the default dispatch since we don't need the rest of
        the stack.
        '''
        if frame.type() == HeartbeatFrame.type():
            self.send_heartbeat()

        elif frame.type() == MethodFrame.type():
            if frame.class_id == 10:
                cb = self._method_map.get(frame.method_id)
                if cb:
                    method = self.clear_synchronous_cb(cb)
                    method(frame)
                else:
                    raise Channel.InvalidMethod(
                        "unsupported method %d on channel %d",
                        frame.method_id, self.channel_id)
            else:
                raise Channel.InvalidClass(
                    "class %d is not supported on channel %d",
                    frame.class_id, self.channel_id)

        else:
            raise Frame.InvalidFrameType(
                "frame type %d is not supported on channel %d",
                frame.type(), self.channel_id)

    def close(self, reply_code=0, reply_text='', class_id=0, method_id=0):
        '''
        Close the main connection connection channel.
        '''
        self._send_close()

    def send_heartbeat(self):
        '''
        Send a heartbeat if needed. Tracks last heartbeat send time.
        '''
        # Note that this does not take into account the time that we last
        # sent a frame. Hearbeats are so small the effect should be quite
        # limited. Also note that we're looking for something near to our
        # scheduled interval, because if this is exact, then we'll likely
        # actually send a heartbeat at twice the period, which could cause
        # a broker to kill the connection if the period is large enough. The
        # 90% bound is arbitrary but seems a sensible enough default.
        if self.connection._heartbeat:
            if time.time() >= (self._last_heartbeat_send + 0.9 *
                               self.connection._heartbeat):
                self.send_frame(HeartbeatFrame(self.channel_id))
                self._last_heartbeat_send = time.time()

    def _recv_start(self, method_frame):
        self.connection._closed = False
        self._send_start_ok()

    def _send_start_ok(self):
        '''Send the start_ok message.'''
        args = Writer()
        args.write_table(self.connection._properties)
        args.write_shortstr(self.connection._login_method)
        args.write_longstr(self.connection._login_response)
        args.write_shortstr(self.connection._locale)
        self.send_frame(MethodFrame(self.channel_id, 10, 11, args))

        self.add_synchronous_cb(self._recv_tune)

    def _recv_tune(self, method_frame):
        self.connection._channel_max = method_frame.args.read_short(
        ) or self.connection._channel_max
        self.connection._frame_max = method_frame.args.read_long(
        ) or self.connection._frame_max

        # Note that 'is' test is required here, as 0 and None are distinct
        if self.connection._heartbeat is None:
            self.connection._heartbeat = method_frame.args.read_short()

        self._send_tune_ok()
        self._send_open()

        # 4.2.7: The client should start sending heartbeats after receiving a
        # Connection.Tune method
        self.send_heartbeat()

    def _send_tune_ok(self):
        args = Writer()
        args.write_short(self.connection._channel_max)
        args.write_long(self.connection._frame_max)

        if self.connection._heartbeat:
            args.write_short(self.connection._heartbeat)
        else:
            args.write_short(0)

        self.send_frame(MethodFrame(self.channel_id, 10, 31, args))

    def _recv_secure(self, method_frame):
        self._send_open()

    def _send_open(self):
        args = Writer()
        args.write_shortstr(self.connection._vhost)
        args.write_shortstr('')
        args.write_bit(True)  # insist flag for older amqp, not used in 0.9.1

        self.send_frame(MethodFrame(self.channel_id, 10, 40, args))
        self.add_synchronous_cb(self._recv_open_ok)

    def _recv_open_ok(self, method_frame):
        self.connection._connected = True
        self.connection._flush_buffered_frames()
        self.connection._callback_open()

    def _send_close(self):
        args = Writer()
        args.write_short(self.connection._close_info['reply_code'])
        args.write_shortstr(self.connection._close_info['reply_text'][:255])
        args.write_short(self.connection._close_info['class_id'])
        args.write_short(self.connection._close_info['method_id'])
        self.send_frame(MethodFrame(self.channel_id, 10, 50, args))
        self.add_synchronous_cb(self._recv_close_ok)

    def _recv_close(self, method_frame):
        self.connection._close_info = {
            'reply_code': method_frame.args.read_short(),
            'reply_text': method_frame.args.read_shortstr(),
            'class_id': method_frame.args.read_short(),
            'method_id': method_frame.args.read_short()
        }

        # TODO: wait to disconnect until the close_ok has been flushed, but
        # that's a pain
        self._send_close_ok()

        self.connection._closed = True
        self.connection.disconnect()
        self.connection._callback_close()

    def _send_close_ok(self):
        self.send_frame(MethodFrame(self.channel_id, 10, 51))

    def _recv_close_ok(self, method_frame):
        self.connection._closed = True
        self.connection.disconnect()
        self.connection._callback_close()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

import warnings

from haigha.transports.transport import Transport

try:
    from eventsocket import EventSocket
    import event
except ImportError:
    warnings.warn('Failed to load EventSocket and event modules')
    EventSocket = None
    event = None


class EventTransport(Transport):

    '''
    Transport using libevent-based EventSocket.
    '''

    def __init__(self, *args):
        super(EventTransport, self).__init__(*args)
        self._synchronous = False

    ###
    # EventSocket callbacks
    ###
    def _sock_close_cb(self, sock):
        self._connection.transport_closed(
            msg='socket to %s closed unexpectedly' % (self._host),
        )

    def _sock_error_cb(self, sock, msg, exception=None):
        self._connection.transport_closed(
            msg='error on connection to %s: %s' % (self._host, msg)
        )

    def _sock_read_cb(self, sock):
        self.connection.read_frames()

    ###
    # Transport API
    ###
    def connect(self, (host, port)):
        '''
        Connect assuming a host and port tuple. Implemented as non-blocking,
        and will close the transport if there's an error
        '''
        self._host = "%s:%s" % (host, port)
        self._sock = EventSocket(
            read_cb=self._sock_read_cb,
            close_cb=self._sock_close_cb,
            error_cb=self._sock_error_cb,
            debug=self.connection.debug,
            logger=self.connection.logger)
        if self.connection._sock_opts:
            for k, v in self.connection._sock_opts.iteritems():
                family, type = k
                self._sock.setsockopt(family, type, v)
        self._sock.setblocking(False)
        self._sock.connect(
            (host, port), timeout=self.connection._connect_timeout)
        self._heartbeat_timeout = None

    def read(self, timeout=None):
        '''
        Read from the transport. If no data is available, should return None.
        The timeout is ignored as this returns only data that has already
        been buffered locally.
        '''
        # NOTE: copying over this comment from Connection, because there is
        # knowledge captured here, even if the details are stale
        # Because of the timer callback to dataRead when we re-buffered,
        # there's a chance that in between we've lost the socket. If that's
        # the case, just silently return as some code elsewhere would have
        # already notified us. That bug could be fixed by improving the
        # message reading so that we consume all possible messages and ensure
        # that only a partial message was rebuffered, so that we can rely on
        # the next read event to read the subsequent message.
        if not hasattr(self, '_sock'):
            return None

        # This is sort of a hack because we're faking that data is ready, but
        # it works for purposes of supporting timeouts
        if timeout:
            if self._heartbeat_timeout:
                self._heartbeat_timeout.delete()
            self._heartbeat_timeout = \
                event.timeout(timeout, self._sock_read_cb, self._sock)
        elif self._heartbeat_timeout:
            self._heartbeat_timeout.delete()
            self._heartbeat_timeout = None

        return self._sock.read()

    def buffer(self, data):
        '''
        Buffer unused bytes from the input stream.
        '''
        if not hasattr(self, '_sock'):
            return None
        self._sock.buffer(data)

    def write(self, data):
        '''
        Write some bytes to the transport.
        '''
        if not hasattr(self, '_sock'):
            return
        self._sock.write(data)

    def disconnect(self):
        '''
        Disconnect from the transport. Typically socket.close(). This call is
        welcome to raise exceptions, which the Connection will catch.

        The transport is encouraged to allow for any pending writes to complete
        before closing the socket.
        '''
        if not hasattr(self, '_sock'):
            return

        # TODO: If there are bytes left on the output, queue the close for
        # later.
        self._sock.close_cb = None
        self._sock.close()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.frames.frame import Frame
from haigha.reader import Reader
from haigha.writer import Writer


class MethodFrame(Frame):

    '''
    Frame which carries identifier for methods.
    '''

    @classmethod
    def type(cls):
        return 1

    @property
    def class_id(self):
        return self._class_id

    @property
    def method_id(self):
        return self._method_id

    @property
    def args(self):
        return self._args

    @classmethod
    def parse(self, channel_id, payload):
        class_id = payload.read_short()
        method_id = payload.read_short()
        return MethodFrame(channel_id, class_id, method_id, payload)

    def __init__(self, channel_id, class_id, method_id, args=None):
        Frame.__init__(self, channel_id)
        self._class_id = class_id
        self._method_id = method_id
        self._args = args

    def __str__(self):
        if isinstance(self.args, (Reader, Writer)):
            return "%s[channel: %d, class_id: %d, method_id: %d, args: %s]" %\
                (self.__class__.__name__, self.channel_id,
                 self.class_id, self.method_id, str(self.args))
        else:
            return "%s[channel: %d, class_id: %d, method_id: %d, args: None]" %\
                (self.__class__.__name__, self.channel_id,
                 self.class_id, self.method_id)

    def write_frame(self, buf):
        writer = Writer(buf)
        writer.write_octet(self.type())
        writer.write_short(self.channel_id)

        # Write a temporary value for the total length of the frame
        stream_args_len_pos = len(buf)
        writer.write_long(0)

        # Mark the point in the stream where we start writing arguments,
        # *including* the class and method ids.
        stream_method_pos = len(buf)

        writer.write_short(self.class_id)
        writer.write_short(self.method_id)

        # This is assuming that args is a Writer
        if self._args is not None:
            writer.write(self._args.buffer())

        # Write the total length back at the position we allocated
        stream_len = len(buf) - stream_method_pos
        writer.write_long_at(stream_len, stream_args_len_pos)

        # Write the footer
        writer.write_octet(0xce)


MethodFrame.register()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

from haigha.classes.protocol_class import ProtocolClass
from haigha.frames.frame import Frame
from haigha.frames.content_frame import ContentFrame
from haigha.frames.header_frame import HeaderFrame
from haigha.exceptions import ChannelError, ChannelClosed, ConnectionClosed

# Defined here so it's easier to test


class SyncWrapper(object):

    def __init__(self, cb):
        self._cb = cb
        self._read = True
        self._result = None

    def __eq__(self, other):
        return other == self._cb or \
            (isinstance(other, SyncWrapper) and other._cb == self._cb)

    def __call__(self, *args, **kwargs):
        self._read = False
        self._result = self._cb(*args, **kwargs)


class Channel(object):

    '''
    Define a channel
    '''

    class InvalidClass(ChannelError):
        '''The method frame referenced an invalid class.  Non-fatal.'''

    class InvalidMethod(ChannelError):
        '''The method frame referenced an invalid method.  Non-fatal.'''

    class Inactive(ChannelError):
        '''
        Tried to send a content frame while the channel was inactive.
        Non-fatal.
        '''

    def __init__(self, connection, channel_id, class_map, **kwargs):
        '''
        Initialize with a handle to the connection and an id. Caller must
        supply a mapping of {class_id:ProtocolClass} which defines what
        classes and methods this channel will support.
        '''
        self._connection = connection
        self._channel_id = channel_id

        self._class_map = {}
        for _id, _class in class_map.iteritems():
            impl = _class(self)
            setattr(self, impl.name, impl)
            self._class_map[_id] = impl

        # Out-bound mix of pending frames and synchronous callbacks
        self._pending_events = deque()

        # Incoming frame buffer
        self._frame_buffer = deque()

        # Listeners for when channel opens
        self._open_listeners = set()

        # Listeners for when channel closes
        self._close_listeners = set()

        # Moving state out of protocol class so that it's accessible even
        # after we've closed and deleted references to the protocol classes.
        # Note though that many of these fields are written to directly
        # from within ChannelClass.
        self._closed = False
        self._close_info = {
            'reply_code': 0,
            'reply_text': 'first connect',
            'class_id': 0,
            'method_id': 0
        }
        self._active = True

        self._synchronous = kwargs.get('synchronous', False)

    @property
    def connection(self):
        return self._connection

    @property
    def channel_id(self):
        return self._channel_id

    @property
    def logger(self):
        '''Return a shared logger handle for the channel.'''
        return self._connection.logger

    @property
    def closed(self):
        '''Return whether this channel has been closed.'''
        return self._closed

    @property
    def close_info(self):
        '''Return dict with information on why this channel is closed.  Will
        return None if the channel is open.'''
        return self._close_info if self._closed else None

    @property
    def active(self):
        '''
        Return True if flow control turned off, False if flow control is on.
        '''
        return self._active

    @property
    def synchronous(self):
        '''
        Return if this channel is acting synchronous, of its own accord or
        because the connection is synchronous.
        '''
        return self._synchronous or self._connection.synchronous

    def add_open_listener(self, listener):
        '''
        Add a listener for open events on this channel. The listener should be
        a callable that can take one argument, the channel that is opened.
        Listeners will not be called in any particular order.
        '''
        self._open_listeners.add(listener)

    def remove_open_listener(self, listener):
        '''
        Remove an open event listener. Will do nothing if the listener is not
        registered.
        '''
        self._open_listeners.discard(listener)

    def _notify_open_listeners(self):
        '''Call all the open listeners.'''
        for listener in self._open_listeners:
            listener(self)

    def add_close_listener(self, listener):
        '''
        Add a listener for close events on this channel. The listener should be
        a callable that can take one argument, the channel that is closed.
        Listeners will not be called in any particular order.
        '''
        self._close_listeners.add(listener)

    def remove_close_listener(self, listener):
        '''
        Remove a close event listener. Will do nothing if the listener is not
        registered.
        '''
        self._close_listeners.discard(listener)

    def _notify_close_listeners(self):
        '''Call all the close listeners.'''
        for listener in self._close_listeners:
            listener(self)

    def open(self):
        '''
        Open this channel.  Routes to channel.open.
        '''
        self.channel.open()

    def close(self, reply_code=0, reply_text='', class_id=0, method_id=0):
        '''
        Close this channel.  Routes to channel.close.
        '''
        # In the off chance that we call this twice. A good example is if
        # there's an error in close listeners and so we're still inside a
        # single call to process_frames, which will try to close this channel
        # if there's an exception.
        if hasattr(self, 'channel'):
            self.channel.close(reply_code, reply_text, class_id, method_id)

    def publish(self, *args, **kwargs):
        '''
        Standard publish.  See basic.publish.
        '''
        self.basic.publish(*args, **kwargs)

    def publish_synchronous(self, *args, **kwargs):
        '''
        Helper for publishing a message using transactions.  If 'cb' keyword
        arg is supplied, will be called when the transaction is committed.
        '''
        cb = kwargs.pop('cb', None)
        self.tx.select()
        self.basic.publish(*args, **kwargs)
        self.tx.commit(cb=cb)

    def dispatch(self, method_frame):
        '''
        Dispatch a method.
        '''
        klass = self._class_map.get(method_frame.class_id)
        if klass:
            klass.dispatch(method_frame)
        else:
            raise Channel.InvalidClass(
                "class %d is not supported on channel %d",
                method_frame.class_id, self.channel_id)

    def buffer_frame(self, frame):
        '''
        Buffer an input frame.  Will append to current list of frames and
        ensure there's a pending event to process the queue.
        '''
        self._frame_buffer.append(frame)

    def process_frames(self):
        '''
        Process the input buffer.
        '''
        while len(self._frame_buffer):
            try:
                # It would make sense to call next_frame, but it's
                # technically faster to repeat the code here.
                frame = self._frame_buffer.popleft()
                self.dispatch(frame)
            except ProtocolClass.FrameUnderflow:
                return
            except (ConnectionClosed, ChannelClosed):
                # Immediately raise if connection or channel is closed
                raise
            except Exception:
                # Spec says that channel should be closed if there's a framing
                # error. Unsure if we can send close if the current exception
                # is transport level (e.g. gevent.GreenletExit)
                try:
                    self.close(500, "Failed to dispatch %s" % (str(frame)))
                finally:
                    raise

    def next_frame(self):
        '''
        Pop the next frame off the input queue. If the queue is empty, will
        return None.
        '''
        if len(self._frame_buffer):
            return self._frame_buffer.popleft()
        return None

    def requeue_frames(self, frames):
        '''
        Requeue a list of frames. Will append to the head of the frame buffer.
        Frames should be in reverse order. Really only used to support
        BasicClass content consumers.
        '''
        self._frame_buffer.extendleft(frames)

    def send_frame(self, frame):
        '''
        Queue a frame for sending.  Will send immediately if there are no
        pending synchronous transactions on this connection.
        '''
        if self.closed:
            if self.close_info and len(self.close_info['reply_text']) > 0:
                raise ChannelClosed(
                    "channel %d is closed: %s : %s",
                    self.channel_id,
                    self.close_info['reply_code'],
                    self.close_info['reply_text'])
            raise ChannelClosed()

        # If there's any pending event at all, then it means that when the
        # current dispatch loop started, all possible frames were flushed
        # and the remaining item(s) starts with a sync callback. After careful
        # consideration, it seems that it's safe to assume the len>0 means to
        # buffer the frame. The other advantage here is
        if not len(self._pending_events):
            if not self._active and \
                    isinstance(frame, (ContentFrame, HeaderFrame)):
                raise Channel.Inactive(
                    "Channel %d flow control activated", self.channel_id)
            self._connection.send_frame(frame)
        else:
            self._pending_events.append(frame)

    def add_synchronous_cb(self, cb):
        '''
        Add an expectation of a callback to release a synchronous transaction.
        '''
        if self.connection.synchronous or self._synchronous:
            wrapper = SyncWrapper(cb)
            self._pending_events.append(wrapper)
            while wrapper._read:
                # Don't check that the channel has been closed until after
                # reading frames, in the case that this is processing a clean
                # channel closed. If there's a protocol error during
                # read_frames, this will loop back around and result in a
                # channel closed exception.
                if self.closed:
                    if self.close_info and \
                            len(self.close_info['reply_text']) > 0:
                        raise ChannelClosed(
                            "channel %d is closed: %s : %s",
                            self.channel_id,
                            self.close_info['reply_code'],
                            self.close_info['reply_text'])
                    raise ChannelClosed()
                self.connection.read_frames()

            return wrapper._result
        else:
            self._pending_events.append(cb)

    def clear_synchronous_cb(self, cb):
        '''
        If the callback is the current expected callback, will clear it off the
        stack.  Else will raise in exception if there's an expectation but this
        doesn't satisfy it.
        '''
        if len(self._pending_events):
            ev = self._pending_events[0]

            # We can't have a strict check using this simple mechanism,
            # because we could be waiting for a synch response while messages
            # are being published. So for now, if it's not in the list, do a
            # check to see if the callback is in the pending list, and if so,
            # then raise, because it means we received stuff out of order.
            # Else just pass it through. Note that this situation could happen
            # on any broker-initiated message.
            if ev == cb:
                self._pending_events.popleft()
                self._flush_pending_events()
                return ev

            elif cb in self._pending_events:
                raise ChannelError(
                    "Expected synchronous callback %s, got %s", ev, cb)
        # Return the passed-in callback by default
        return cb

    def _flush_pending_events(self):
        '''
        Send pending frames that are in the event queue.
        '''
        while len(self._pending_events) and \
                isinstance(self._pending_events[0], Frame):
            self._connection.send_frame(self._pending_events.popleft())

    def _closed_cb(self, final_frame=None):
        '''
        "Private" callback from the ChannelClass when a channel is closed. Only
        called after broker initiated close, or we receive a close_ok. Caller
        has the option to send a final frame, to be used to bypass any
        synchronous or otherwise-pending frames so that the channel can be
        cleanly closed.
        '''
        # delete all pending data and send final frame if thre is one. note
        # that it bypasses send_frame so that even if the closed state is set,
        # the frame is published.
        if final_frame:
            self._connection.send_frame(final_frame)

        try:
            self._notify_close_listeners()
        finally:
            self._pending_events = deque()
            self._frame_buffer = deque()

            # clear out other references for faster cleanup
            for protocol_class in self._class_map.values():
                protocol_class._cleanup()
                delattr(self, protocol_class.name)
            self._connection = None
            self._class_map = None
            self._close_listeners = set()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.writer import Writer
from haigha.frames.frame import Frame


class ContentFrame(Frame):

    '''
    Frame for reading in content.
    '''

    @classmethod
    def type(cls):
        return 3

    @property
    def payload(self):
        return self._payload

    @classmethod
    def parse(self, channel_id, payload):
        return ContentFrame(channel_id, payload)

    @classmethod
    def create_frames(self, channel_id, buf, frame_max):
        '''
        A generator which will create frames from a buffer given a max
        frame size.
        '''
        size = frame_max - 8   # 8 bytes overhead for frame header and footer
        offset = 0
        while True:
            payload = buf[offset:(offset + size)]
            if len(payload) == 0:
                break
            offset += size

            yield ContentFrame(channel_id, payload)
            if offset >= len(buf):
                break

    def __init__(self, channel_id, payload):
        Frame.__init__(self, channel_id)
        self._payload = payload

    def __str__(self):
        if isinstance(self._payload, str):
            payload = ''.join(['\\x%s' % (c.encode('hex'))
                               for c in self._payload])
        else:
            payload = str(self._payload)

        return "%s[channel: %d, payload: %s]" % (
            self.__class__.__name__, self.channel_id, payload)

    def write_frame(self, buf):
        '''
        Write the frame into an existing buffer.
        '''
        writer = Writer(buf)

        writer.write_octet(self.type()).\
            write_short(self.channel_id).\
            write_long(len(self._payload)).\
            write(self._payload).\
            write_octet(0xce)


ContentFrame.register()
